#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from graphdot.codegen.typetool import common_concrete_type, common_min_type


class Series(np.ndarray):
    '''A thin wrapper to customize serialization behavior'''
    def __new__(cls, input):
        if isinstance(input, np.ndarray):
            series = input.view(cls)
            if np.issctype(series.dtype):
                series._concrete_type = series.dtype
            else:
                series._concrete_type = common_concrete_type.of_values(input)
        else:
            t = common_min_type.of_values(input)
            dtype = np.dtype(t) if np.issctype(t) else np.dtype(np.object)
            series = np.empty(len(input), dtype=dtype).view(cls)  # ensures 1D
            series[:] = input
            series._concrete_type = t
        return series

    def __repr__(self):
        return np.array2string(self, separator=',', max_line_width=1e20)

    @property
    def concrete_type(self):
        return self._concrete_type

    def __reduce__(self):
        recon, args, state = super(Series, self).__reduce__()
        return (recon, args, (state, self.__dict__))

    def __setstate__(self, states):
        state, dict_ = states
        self.__dict__.update(**dict_)
        super(Series, self).__setstate__(state)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import os
import copy
from contextlib import contextmanager


class Template:
    """A template string with placeholders in the format of ${...}, which
    can be rendered on-demand. Helpful for code generation.

    Parameters
    ----------
    template: str
        A template string containing placeholders in the format of ${...}.
    escape: bool
        Whether or not backslash (\\) should be escaped in the fill-in contents
        to the placeholders.
    """

    def __init__(self, template, escape=True):
        if os.path.isfile(template):
            self.template = open(template).read()
        else:
            self.template = template
        self.escape = escape

    @contextmanager
    def context(self, **kwargs):
        t = copy.copy(self)
        t.template = re.sub(
            r'\?{([^}]+)}',
            lambda m: 'true' if eval(m.group(1), kwargs) else 'false',
            t.template
        )
        yield t

    def render(self, **substitutions):
        """Substitute placeholders using the syntax symbol=replacement.
        Partial renderings are allowed. If replacement is list-like, use the
        trailing sequence of symbol match in placeholders to join members;
        otherwise do plain substitution.
        """

        text = self.template
        for symbol in sorted(substitutions, key=lambda s: (-len(s), s)):
            repl = substitutions[symbol]
            if isinstance(repl, (list, tuple)):
                pattern = r'\${%s([^}]*)}' % symbol
                repl = [str(r) for r in repl]
                text = re.sub(pattern, lambda m: m.group(1).join(repl), text)
            else:
                pattern = r'\${%s}' % symbol
                repl = str(repl)
                if self.escape is False:
                    repl = repl.replace('\\', r'\\')
                text = re.sub(pattern, repl, text)
        return text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from scipy.optimize import minimize
from ._chol_solver import CholSolver


class GaussianProcessRegressor:
    """Gaussian process regression (GPR).

    Parameters
    ----------
    kernel: kernel instance
        The covariance function of the GP.
    alpha: float > 0, default = 1e-10
        Value added to the diagonal of the kernel matrix during fitting. Larger
        values correspond to increased noise level in the observations. A
        practical usage of this parameter is to prevent potential numerical
        stability issues during fitting, and ensures that the kernel matrix is
        always positive definite in the precense of duplicate entries and/or
        round-off error.
    optimizer: one of (str, True, None, callable)
        A string or callable that represents one of the optimizers usable in
        the scipy.optimize.minimize method.
        If None, no hyperparameter optimization will be carried out in fitting.
        If True, the optimizer will default to L-BFGS-B.
    normalize_y: boolean
        Whether to normalize the target values y so that the mean and variance
        become 0 and 1, respectively. Recommended for cases where zero-mean,
        unit-variance kernels are used. Normalisation will be reversed before
        the GP predictions are returned.
    """

    def __init__(self, kernel, alpha=1e-10, optimizer=None, normalize_y=False):
        self.kernel = kernel
        self.alpha = alpha
        self.optimizer = optimizer
        if optimizer is True:
            self.optimizer = 'L-BFGS-B'
        self.normalize_y = normalize_y

    def fit(self, X, y):
        """Train a GPR model with the given data.

        Parameters
        ----------
        X: list of objects or feature vectors.
            Input values of the training data.
        y: 1D array
            Output/target values of the training data.

        Returns
        -------
        self: returns an instance of self.
        """
        # set X
        self.X = X
        # set y
        if self.normalize_y is True:
            self.y_mean, self.y_std = np.mean(y), np.std(y)
            self.y = (y - self.y_mean) / self.y_std
        else:
            self.y_mean, self.y_std = 0, 1
            self.y = np.copy(y)
        # train model
        if self.optimizer:
            self._optimize_hyperparameters(X, y)
        self.K = K = self.kernel(X)
        K.flat[::len(K) + 1] += self.alpha
        self.L = CholSolver(K)
        self.Ky = self.L(y)
        return self

    def predict(self, Z, return_std=False, return_cov=False):
        """Predict using the trained GPR model.

        Parameters
        ----------
        Z: list of objects or feature vectors.
            Input values of the unknown data.
        return_std: boolean
            If True, the standard-deviations of the predictions at the query
            points are returned along with the mean.
        return_cov: boolean
            If True, the covariance of the predictions at the query points are
            returned along with the mean.

        Returns
        -------
        ymean: 1D array
            Mean of the predictive distribution at query points.
        std: 1D array
            Standard deviation of the predictive distribution at query points.
        cov: 2D matrix
            Covariance of the predictive distribution at query points.
        """
        if not hasattr(self, 'K'):
            raise RuntimeError('Model not trained.')
        Ks = self.kernel(Z, self.X)
        ymean = (Ks @ self.Ky) * self.y_std + self.y_mean
        if return_std is True:
            Kss = self.kernel.diag(Z)
            std = np.sqrt(np.maximum(Kss - (Ks @ self.L(Kss.T)).diagonal(), 0))
            return (ymean, std * self.y_std)
        elif return_cov is True:
            Kss = self.kernel(Z)
            cov = np.maximum(Kss - Ks @ self.L(Kss.T), 0)
            return (ymean, cov * self.y_std**2)
        else:
            return ymean

    def log_marginal_likelihood(self, theta=None, eval_gradient=False,
                                clone_kernel=True, verbose=False):
        """Returns log-marginal likelihood of a given set of log-scale
        hyperparameters for the training data as previously given by fit().

        Parameters
        ----------
        theta: array-like
            Kernel hyperparameters for which the log-marginal likelihood is
            to be evaluated. If None, the current hyperparameters will be used.
        eval_gradient: boolean
            If True, the gradient of the log-marginal likelihood with respect
            to the kernel hyperparameters at position theta will be returned
            alongside.
        clone_kernel: boolean
            If True, the kernel is copied so that probing with theta does not
            alter the trained kernel. If False, the kernel hyperparameters will
            be modified in-place.
        verbose: boolean
            If True, the log-likelihood value and its components will be
            printed to the screen.

        Returns
        -------
        log_likelihood: float
            Log-marginal likelihood of theta for training data.
        log_likelihood_gradient: 1D array
            Gradient of the log-marginal likelihood with respect to the kernel
            hyperparameters at position theta. Only returned when eval_gradient
            is True.
        """
        if theta is None:
            theta = self.kernel.theta

        if clone_kernel is True:
            kernel = self.kernel.clone_with_theta(theta)
        else:
            kernel = self.kernel
            kernel.theta = theta

        if eval_gradient is True:
            K, dK = kernel(self.X, eval_gradient=True)
        else:
            K = kernel(self.X)
        K.flat[::len(K) + 1] += self.alpha

        L = CholSolver(K)
        Ky = L(self.y)
        yKy = self.y @ Ky
        logdet = np.prod(np.linalg.slogdet(K))
        if verbose:
            print(
                'L %.5f = %.5f + %.5f' % (yKy + logdet, yKy, logdet),
                np.exp(theta)
            )

        if eval_gradient is True:
            D_theta = np.zeros(dK.shape[0])
            for i in range(dK.shape[0]):
                dk = dK[i, :, :]
                D_theta[i] = (L(dk).trace() - Ky @ dk @ Ky) * np.exp(theta[i])
            return yKy + logdet, D_theta
        else:
            return yKy + logdet

    def _optimize_hyperparameters(self, X, y):
        opt = minimize(
            fun=lambda theta, self=self: self.log_marginal_likelihood(
                theta, eval_gradient=True, clone_kernel=False
            ),
            method=self.optimizer,
            x0=self.kernel.theta,
            bounds=self.kernel.bounds,
            jac=True,
            tol=1e-3,
        )

        if opt.success:
            self.kernel.theta = opt.x
        else:
            raise RuntimeError(
                f'Hyperparameter optimization did not converge, result:\n{opt}'
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import copy
import functools
import numpy as np
from mendeleev import get_table
from graphdot.graph.adjacency.euclidean import Tent


def copying_lru_cache(*args, **kwargs):

    def decorator(f):
        cached_func = functools.lru_cache(*args, **kwargs)(f)

        def wrapper(*args, **kwargs):
            return copy.deepcopy(cached_func(*args, **kwargs))
        return wrapper
    return decorator


@copying_lru_cache(maxsize=32)
def get_ptable():
    return get_table('elements')


@copying_lru_cache(maxsize=128)
def get_length_scales(name):
    ptable = get_ptable()
    length = np.zeros(ptable.atomic_number.max() + 1)
    for n, r in ptable[['atomic_number', name]].itertuples(False):
        length[n] = r * 0.01  # pm to A
    return length


class AtomicAdjacency:
    """Converts interatomic distances into edge weights

    Parameters
    ----------
    shape: str or function
        A 1D weight real-valued function that converts distance to weight.
    zoom: float
        A zooming factor to be multiplied with inferred length scales.
    images: None or list of 3D vectors
        Displacement vector of periodic images to be checked for finding
        the closest neighbors in a crystal.
    """

    def __init__(self, shape='tent1', length_scale='vdw_radius', zoom=1.0):
        if isinstance(shape, str):
            m = re.match(r'tent([1-9]\d*)', shape)
            if m:
                self.shape = Tent(ord=int(m.group(1)))
            else:
                # raise ValueError(f'Invalid shape: {shape}')
                raise ValueError('Invalid shape: {}'.format(shape))
        else:
            self.shape = shape
        if isinstance(length_scale, str):
            self.ltable = get_length_scales(length_scale)
        else:
            ptbl = get_ptable()
            self.ltable = length_scale * np.ones(ptbl.atomic_number.max() + 1)
        self.ltable *= zoom

    def __call__(self, n1, n2, r):
        """compute adjacency between atoms

        Parameters
        ----------
        n1: int
            Atomic number of the element
        n2: int
            Same as n1
        r: float
            Distance between the two atoms

        Returns
        -------
        float:
            A non-negative weight
        """
        r1 = self.ltable[n1]
        r2 = self.ltable[n2]
        weight = self.shape(r, np.sqrt(r1 * r2))
        return weight

    def cutoff(self, elements):
        max_length_scale = self.ltable[elements].max()
        return self.shape.cutoff(max_length_scale)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Convert spatial distances between nodes to edge adjacencies
"""
import numpy as np
# import numba as nb


# @nb.jitclass([('half_inv_sigma2', nb.float32)])
class Gaussian:
    def __init__(self, h):
        self.half_inv_h2 = 0.5 / h**2

    def compute(self, d):
        return np.exp(d**2 * self.half_inv_h2)


class Tent:
    def __init__(self, ord):
        self.ord = ord

    def __call__(self, d, length_scale):
        s = 1 - d / self.cutoff(length_scale)
        return s ** self.ord if s >= 0 else 0

    def cutoff(self, length_scale):
        return length_scale * 3


if __name__ == '__main__':

    g = Gaussian(1.0)
    print(g(1.0))
    print(g(2.0))
    print(g(3.0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from pycuda import gpuarray
from graphdot.codegen.cpptool import cpptype


@cpptype(p_buffer=np.uintp, capacity=np.int64)
class BlockScratch:
    def __init__(self, capacity):
        self.capacity = ((capacity + 15) // 16) * 16
        if capacity <= 0:
            raise ValueError('Scratch size must be greater than zero.')
        self.buffer = gpuarray.empty(int(self.capacity) * 5, np.float32)

    @property
    def p_buffer(self):
        return self.buffer.ptr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module defines base kernels and composibility rules for creating vertex
and edge kernels for the marginalized graph kernel.
"""
from collections import namedtuple, OrderedDict
import numpy as np
import sympy as sy
from sympy.utilities.lambdify import lambdify
from graphdot.codegen import Template
from graphdot.codegen.cpptool import cpptype
from graphdot.codegen.sympy_printer import cudacxxcode

__all__ = ['BaseKernel',
           'Constant',
           'KroneckerDelta',
           'SquareExponential',
           'RationalQuadratic',
           'Normalize',
           'Compose',
           'TensorProduct',
           'Additive',
           'Convolution',
           ]


class BaseKernel:

    @staticmethod
    def create(kernel, desc, expr, vars, *hyperparameter_specs):

        '''parse expression'''
        if isinstance(expr, str):
            expr = sy.sympify(expr)

        '''check input variables'''
        if len(vars) != 2:
            raise ValueError('A kernel must have exactly two variables')
        vars = [sy.Symbol(v) if isinstance(v, str) else v for v in vars]

        '''parse the list of hyperparameters'''
        hyperdefs = OrderedDict()
        for spec in hyperparameter_specs:
            if not hasattr(spec, '__iter__'):
                symbol = spec
                hyperdefs[symbol] = dict(dtype=np.dtype(np.float32))
            if len(spec) == 1:
                symbol = spec[0]
                hyperdefs[symbol] = dict(dtype=np.dtype(np.float32))
            if len(spec) == 2:
                symbol, dtype = spec
                hyperdefs[symbol] = dict(dtype=np.dtype(dtype))
            if len(spec) == 3:
                symbol, dtype, doc = spec
                hyperdefs[symbol] = dict(dtype=np.dtype(dtype), doc=doc)
            elif len(spec) == 4:
                symbol, dtype, lb, ub = spec
                hyperdefs[symbol] = dict(dtype=np.dtype(dtype),
                                         bounds=(lb, ub))
            elif len(spec) == 5:
                symbol, dtype, lb, ub, doc = spec
                hyperdefs[symbol] = dict(dtype=np.dtype(dtype),
                                         bounds=(lb, ub),
                                         doc=doc)
            else:
                raise ValueError(
                    'Invalid hyperparameter specification, '
                    'must be one of\n'
                    '(symbol)\n',
                    '(symbol, dtype)\n',
                    '(symbol, dtype, doc)\n',
                    '(symbol, dtype, lb, ub)\n',
                    '(symbol, dtype, lb, ub, doc)\n',
                )

        '''create kernel class'''
        class CppType(type):
            @property
            def dtype(cls):
                return cls._dtype

        class Kernel(BaseKernel, metaclass=CppType):

            __name__ = kernel

            _expr = expr
            _vars = vars
            _hyperdefs = hyperdefs
            _dtype = np.dtype([(k, v['dtype']) for k, v in hyperdefs.items()],
                              align=True)

            def __init__(self, *args, **kwargs):

                self._theta_values = values = OrderedDict()
                self._theta_bounds = bounds = OrderedDict()

                for symbol, value in zip(self._hyperdefs, args):
                    values[symbol] = value

                for symbol in self._hyperdefs:
                    try:
                        values[symbol] = kwargs[symbol]
                    except KeyError:
                        if symbol not in values:
                            raise KeyError(
                                f'Hyperparameter {symbol} not provided '
                                f'for {self.__name__}'
                            )

                    try:
                        bounds[symbol] = kwargs['%s_bounds' % symbol]
                    except KeyError:
                        try:
                            bounds[symbol] = self._hyperdefs[symbol]['bounds']
                        except KeyError:
                            raise KeyError(
                                f'Bounds for hyperparameter {symbol} of '
                                f'kernel {self.__name__} not set, and no '
                            )

            # @cached_property
            @property
            def _vars_and_hypers(self):
                if not hasattr(self, '_vars_and_hypers_cached'):
                    self._vars_and_hypers_cached = [
                        *self._vars, *self._hyperdefs.keys()
                    ]
                return self._vars_and_hypers_cached

            # @cached_property
            @property
            def _fun(self):
                if not hasattr(self, '_fun_cached'):
                    self._fun_cached = lambdify(
                        self._vars_and_hypers,
                        self._expr
                    )
                return self._fun_cached
                # return lambdify(self._vars_and_hypers, self._expr)

            # @cached_property
            @property
            def _jac(self):
                if not hasattr(self, '_jac_cached'):
                    self._jac_cached = [
                        lambdify(self._vars_and_hypers, sy.diff(expr, h))
                        for h in self._hyperdefs
                    ]
                return self._jac_cached
                # return [lambdify(self._vars_and_hypers, sy.diff(expr, h))
                #         for h in self._hyperdefs]

            def __call__(self, x1, x2, jac=False):
                if jac is True:
                    return (
                        self._fun(x1, x2, *self.theta),
                        np.array([j(x1, x2, *self.theta) for j in self._jac])
                    )
                else:
                    return self._fun(x1, x2, *self.theta)

            def __repr__(self):
                return Template('${cls}(${theta, }, ${bounds, })').render(
                    cls=self.__name__,
                    theta=[f'{n}={v}' for n, v in self._theta_values.items()],
                    bounds=[f'{n}_bounds={v}'
                            for n, v in self._theta_bounds.items()]
                )

            def gen_expr(self, x, y, jac=False, theta_scope=''):
                nmap = {
                    str(self._vars[0]): x,
                    str(self._vars[1]): y,
                    **{t: theta_scope + t for t in self._hyperdefs}
                }

                if jac is True:
                    return (
                        cudacxxcode(self._expr, nmap),
                        [cudacxxcode(sy.diff(self._expr, h), nmap)
                         for h in self._hyperdefs]
                    )
                else:
                    return cudacxxcode(self._expr, nmap)

            @property
            def dtype(self):
                return self._dtype

            @property
            def state(self):
                return tuple(self._theta_values.values())

            @property
            def theta(self):
                return namedtuple(
                    self.__name__ + 'Hyperparameters',
                    self._theta_values.keys()
                )(**self._theta_values)

            @theta.setter
            def theta(self, seq):
                assert(len(seq) == len(self._theta_values))
                for theta, value in zip(self._hyperdefs, seq):
                    self._theta_values[theta] = value

            @property
            def bounds(self):
                return tuple(self._theta_bounds.values())

        '''furnish doc strings'''
        param_docs = [
            Template(
                '${name}: ${type}\n'
                '    ${desc\n    }\n'
                '${name}_bounds: pair of ${type}\n'
                '    Lower and upper bounds of `${name}`.'
            ).render(
                name=name,
                type=hdef['dtype'],
                desc=[s.strip() for s in hdef.get('doc', '').split('\n')]
            ) for name, hdef in hyperdefs.items()
        ]

        Kernel.__doc__ = Template(
            '${desc}\n'
            '\n'
            'Parameters\n'
            '----------\n'
            '${param_docs\n}',
            escape=False
        ).render(
            desc='\n'.join([s.strip() for s in desc.split('\n')]),
            param_docs=param_docs
        )

        return Kernel

    def __add__(self, k):
        r"""Implements the additive kernel composition semantics, i.e.
        expression ``k1 + k2`` creates
        :math:`k_+(a, b) = k_1(a, b) + k_2(a, b)`"""
        return KernelOperator.add(
            self,
            k if isinstance(k, BaseKernel) else Constant(k)
        )

    def __radd__(self, k):
        return KernelOperator.add(
            k if isinstance(k, BaseKernel) else Constant(k),
            self
        )

    def __mul__(self, k):
        r"""Implements the multiplicative kernel composition semantics, i.e.
        expression ``k1 * k2`` creates
        :math:`k_\times(a, b) = k_1(a, b) \times k_2(a, b)`"""
        return KernelOperator.mul(
            self,
            k if isinstance(k, BaseKernel) else Constant(k),
        )

    def __rmul__(self, k):
        return KernelOperator.mul(
            k if isinstance(k, BaseKernel) else Constant(k),
            self,
        )


class KernelOperator(BaseKernel):
    def __init__(self, k1, k2):
        self.k1 = k1
        self.k2 = k2

    def __repr__(self):
        return f'{repr(self.k1)} {self.opstr} {repr(self.k2)}'

    @property
    def theta(self):
        return (self.k1.theta, self.k2.theta)

    @theta.setter
    def theta(self, seq):
        self.k1.theta = seq[0]
        self.k2.theta = seq[1]

    @property
    def bounds(self):
        return tuple(self.k1.bounds, self.k2.bounds)

    @staticmethod
    def add(k1, k2):
        @cpptype(k1=k1.dtype, k2=k2.dtype)
        class Add(KernelOperator):

            opstr = '+'

            def __call__(self, i, j, jac=False):
                if jac is True:
                    f1, J1 = self.k1(i, j, True)
                    f2, J2 = self.k2(i, j, True)
                    return (f1 + f2, J1 + J2)
                else:
                    return self.k1(i, j, False) + self.k2(i, j, False)

            def gen_expr(self, x, y, jac=False, theta_prefix=''):
                if jac is True:
                    f1, J1 = self.k1.gen_expr(x, y, True, theta_prefix + 'k1.')
                    f2, J2 = self.k2.gen_expr(x, y, True, theta_prefix + 'k2.')
                    return (f'({f1} + {f2})', J1 + J2)
                else:
                    f1 = self.k1.gen_expr(x, y, False, theta_prefix + 'k1.')
                    f2 = self.k2.gen_expr(x, y, False, theta_prefix + 'k2.')
                    return f'({f1} + {f2})'

        return Add(k1, k2)

    @staticmethod
    def mul(k1, k2):
        @cpptype(k1=k1.dtype, k2=k2.dtype)
        class Mul(KernelOperator):

            opstr = '*'

            def __call__(self, i, j, jac=False):
                if jac is True:
                    f1, J1 = self.k1(i, j, True)
                    f2, J2 = self.k2(i, j, True)
                    return (
                        f1 * f2,
                        np.array(
                            [j1 * f2 for j1 in J1] + [f1 * j2 for j2 in J2]
                        )
                    )
                else:
                    return self.k1(i, j, False) * self.k2(i, j, False)

            def gen_expr(self, x, y, jac=False, theta_prefix=''):
                if jac is True:
                    f1, J1 = self.k1.gen_expr(x, y, True, theta_prefix + 'k1.')
                    f2, J2 = self.k2.gen_expr(x, y, True, theta_prefix + 'k2.')
                    return (
                        f'({f1} * {f2})',
                        [f'({j1} * {f2})' for j1 in J1] +
                        [f'({f1} * {j2})' for j2 in J2]
                    )
                else:
                    f1 = self.k1.gen_expr(x, y, False, theta_prefix + 'k1.')
                    f2 = self.k2.gen_expr(x, y, False, theta_prefix + 'k2.')
                    return f'({f1} * {f2})'

        return Mul(k1, k2)


@cpptype([])
class _Multiply(BaseKernel):
    """
    Direct product is technically not a valid base kernel
    Only used internally for treating weighted edges
    """

    def __call__(self, x1, x2, jac=False):
        if jac is True:
            return x1 * x2, np.array([])
        else:
            return x1 * x2

    def __repr__(self):
        return '_Multiply()'

    def gen_expr(self, x, y, jac=False, theta_prefix=''):
        f = f'({x} * {y})'
        if jac is True:
            return f, []
        else:
            return f

    @property
    def theta(self):
        return tuple()

    @theta.setter
    def theta(self, seq):
        pass

    @property
    def bounds(self):
        return tuple()


def Constant(c, c_bounds=(0, np.inf)):
    r"""Creates a no-op kernel that returns a constant value (often being 1),
    i.e. :math:`k_\mathrm{c}(\cdot, \cdot) \equiv constant`

    Parameters
    ----------
    c: float > 0
        The constant value.

    Returns
    -------
    BaseKernel
        A kernel instance of corresponding behavior
    """

    @cpptype(c=np.float32)
    class ConstantKernel(BaseKernel):
        def __init__(self, c, c_bounds):
            self.c = float(c)
            self.c_bounds = c_bounds

        def __call__(self, i, j, jac=False):
            if jac is True:
                return self.c, np.ones(1)
            else:
                return self.c

        def __repr__(self):
            return f'Constant({self.c})'

        def gen_expr(self, x, y, jac=False, theta_prefix=''):
            f = f'{theta_prefix}c'
            if jac is True:
                return f, ['1.0f']
            else:
                return f

        @property
        def theta(self):
            return namedtuple(
                'ConstantHyperparameters',
                ['c']
            )(self.c)

        @theta.setter
        def theta(self, seq):
            self.c = seq[0]

        @property
        def bounds(self):
            return (self.c_bounds,)

    return ConstantKernel(c, c_bounds)


def KroneckerDelta(h, h_bounds=(1e-3, 1)):
    r"""Creates a Kronecker delta kernel that returns either h or 1 depending
    on whether two objects compare equal, i.e. :math:`k_\delta(i, j) =
    \begin{cases} 1, i = j \\ h, otherwise \end{cases}`

    Parameters
    ----------
    h: float in (0, 1)
        The value of the kernel when two objects do not compare equal

    Returns
    -------
    BaseKernel
        A kernel instance of corresponding behavior
    """

    @cpptype(h=np.float32)
    class KroneckerDeltaKernel(BaseKernel):

        def __init__(self, h, h_bounds):
            self.h = float(h)
            self.h_bounds = h_bounds

        def __call__(self, i, j, jac=False):
            if jac is True:
                return (
                    1.0 if i == j else self.h,
                    np.array([0.0 if i == j else 1.0])
                )
            else:
                return 1.0 if i == j else self.h

        def __repr__(self):
            return f'KroneckerDelta({self.h})'

        def gen_expr(self, x, y, jac=False, theta_prefix=''):
            f = f'({x} == {y} ? 1.0f : {theta_prefix}h)'
            if jac is True:
                return f, [f'({x} == {y} ? 0.0f : 1.0f)']
            else:
                return f

        @property
        def theta(self):
            return namedtuple(
                'KroneckerDeltaHyperparameters',
                ['h']
            )(self.h)

        @theta.setter
        def theta(self, seq):
            self.h = seq[0]

        @property
        def bounds(self):
            return (self.h_bounds,)

    return KroneckerDeltaKernel(h, h_bounds)


SquareExponential = BaseKernel.create(
    'SquareExponential',

    r"""A square exponential kernel smoothly transitions from 1 to
    0 as the distance between two vectors increases from zero to infinity, i.e.
    :math:`k_\mathrm{se}(\mathbf{x}, \mathbf{y}) = \exp(-\frac{1}{2}
    \frac{\lVert \mathbf{x} - \mathbf{y} \rVert^2}{\sigma^2})`""",

    'exp(-0.5 * (x - y)**2 * length_scale**-2)',

    ('x', 'y'),

    ('length_scale', np.float32, 1e-6, np.inf,
     r"""Determines how quickly should the kernel decay to zero. The kernel has
     a value of approx. 0.606 at one length scale, 0.135 at two length
     scales, and 0.011 at three length scales.""")
)

RationalQuadratic = BaseKernel.create(
    'RationalQuadratic',

    r"""A rational quadratic kernel is equivalent to the sum of many square
    exponential kernels with different length scales. The parameter `alpha`
    tunes the relative weights between large and small length scales. When
    alpha approaches infinity, the kernel is identical to the square
    exponential kernel.""",

    '(1 + (x - y)**2 / (2 * alpha * length_scale**2))**(-alpha)',

    ('x', 'y'),

    ('length_scale', np.float32, 1e-6, np.inf,
     r"""The smallest length scale of the square exponential components."""),
    ('alpha', np.float32, 1e-3, np.inf,
     r"""The relative weights of large-scale square exponential components.
     Larger alpha values leads to a faster decay of the weights for larger
     length scales.""")
)


def Normalize(kernel):
    r"""Normalize a positive semidefinite kernel to produce the cosine of angle
    between the input variables.
    :math:`k_{N}(X, Y) = \frac{k_{base}(X, Y)}
                              {\sqrt{k_{base}(X, X) k_{base}(Y, Y)}}`

    Parameters
    ----------
    kernel: base kernel
        An elementary or composite base kernel.
    """

    @cpptype(kernel=kernel.dtype)
    class Normalized(BaseKernel):
        def __init__(self, kernel):
            self.kernel = kernel

        def __call__(self, X, Y, jac=False):
            if jac is True:
                Fxx, Jxx = self.kernel(X, X, jac=True)
                Fxy, Jxy = self.kernel(X, Y, jac=True)
                Fyy, Jyy = self.kernel(Y, Y, jac=True)

                if Fxx > 0 and Fyy > 0:
                    return (
                        Fxy * (Fxx * Fyy)**-0.5,
                        (Jxy * (Fxx * Fyy)**-0.5
                         - (0.5 * Fxy * (Fxx * Fyy)**-1.5
                            * (Jxx * Fyy + Fxx * Jyy)))
                    )
                else:
                    return (0.0, np.zeros_like(Jxy))
            else:
                Fxx = self.kernel(X, X)
                Fxy = self.kernel(X, Y)
                Fyy = self.kernel(Y, Y)
                return Fxy * (Fxx * Fyy)**-0.5 if Fxx > 0 and Fyy > 0 else 0.0

        def __repr__(self):
            return f'Normalize({repr(self.kernel)})'

        def gen_expr(self, x, y, jac=False, theta_prefix=''):
            F, J = self.kernel.gen_expr(
                '_1', '_2', True, theta_prefix + 'kernel.'
            )
            f = Template(
                r'normalize([&](auto _1, auto _2){return ${f};}, ${x}, ${y})'
            ).render(
                x=x, y=y, f=F
            )
            if jac is True:
                template = Template(
                    r'''normalize_jacobian(
                            [&](auto _1, auto _2){return ${f};},
                            [&](auto _1, auto _2){return ${j};},
                            ${x},
                            ${y}
                        )'''
                )
                jacobian = [template.render(x=x, y=y, f=F, j=j) for j in J]
                return f, jacobian
            else:
                return f

        @property
        def theta(self):
            return namedtuple(
                'NormalizeHyperparameters',
                ['kernel']
            )(self.kernel.theta)

        @theta.setter
        def theta(self, seq):
            self.kernel.theta = seq[0]

        @property
        def bounds(self):
            return (self.kernel.bounds,)

    return Normalized(kernel)


def Compose(oper, **kw_kernels):
    r"""Creates a composite base kernel via the usage of a reduction operator
    to combine the outputs of multiple scalar kernels on individual features.
    :math:`k_\mathrm{composite}(X, Y; \mathrm{op}) =
    k_{a_1}(X_{a_1}, Y_{a_1}) \mathrm{op} k_{a_2}(X_{a_2}, Y_{a_2}) \mathrm{op}
    \ldots`

    Parameters
    ----------
    oper: str
        A reduction operator. Due to positive definiteness requirements, the
        available options are currently limited to '+', '*'.
    kw_kernels: dict of attribute=kernel pairs
        The kernels can be any base kernels and their compositions as defined
        in this module, while features should be strings that represent
        valid Python/C++ identifiers.
    """
    oplib = {
        '+': dict(
            ufunc=np.add,                           # associated numpy ufunc
            jfunc=lambda F, f, j: j,                # Jacobian evaluator
            jgen=lambda F_expr, j_expr, i: j_expr,  # Jacobian code generator
        ),
        '*': dict(
            ufunc=np.multiply,
            jfunc=lambda F, f, j: F / f * j,
            jgen=lambda F_expr, j_expr, i: Template('(${X * })').render(
                X=F_expr[:i] + (j_expr,) + F_expr[i + 1:]
            )
        ),
    }

    if oper not in oplib:
        raise ValueError(f'Invalid reduction operator {repr(oper)}.')

    @cpptype([(key, ker.dtype) for key, ker in kw_kernels.items()])
    class Composite(BaseKernel):
        def __init__(self, opstr, ufunc, jfunc, jgen, **kw_kernels):
            self.opstr = opstr
            self.ufunc = ufunc
            self.jfunc = jfunc
            self.jgen = jgen
            self.kw_kernels = kw_kernels

        def __repr__(self):
            return Template('Compose(${opstr}, ${kwexpr, })').render(
                opstr=repr(self.opstr),
                kwexpr=[f'{k}={repr(K)}' for k, K in self.kw_kernels.items()])

        def __call__(self, X, Y, jac=False):
            if jac is True:
                F, J = list(
                    zip(*[kernel(X[key], Y[key], True)
                          for key, kernel in self.kw_kernels.items()])
                )
                S = self.ufunc.reduce(F)
                jacobian = np.array([
                    self.jfunc(S, f, j) for i, f in enumerate(F) for j in J[i]
                ])
                return S, jacobian
            else:
                return self.ufunc.reduce([
                    f(X[k], Y[k]) for k, f in self.kw_kernels.items()
                ])

        def gen_expr(self, x, y, jac=False, theta_prefix=''):
            F, J = list(
                zip(*[kernel.gen_expr('%s.%s' % (x, key),
                                      '%s.%s' % (y, key),
                                      True,
                                      '%s%s.' % (theta_prefix, key))
                      for key, kernel in self.kw_kernels.items()])
            )
            f = Template('(${F ${opstr} })').render(opstr=self.opstr, F=F)
            if jac is True:
                jacobian = [
                    self.jgen(F, j, i) for i, _ in enumerate(F) for j in J[i]
                ]
                return f, jacobian
            else:
                return f

        @property
        def theta(self):
            return namedtuple(
                'CompositeHyperparameters',
                self.kw_kernels.keys()
            )(*[k.theta for k in self.kw_kernels.values()])

        @theta.setter
        def theta(self, seq):
            for kernel, value in zip(self.kw_kernels.values(), seq):
                kernel.theta = value

        @property
        def bounds(self):
            return tuple(k.bounds for k in self.kw_kernels.values())

    # for the .state property of cpptype
    for key in kw_kernels:
        setattr(Composite, key,
                property(lambda self, key=key: self.kw_kernels[key]))

    return Composite(oper, **oplib[oper], **kw_kernels)


def TensorProduct(**kw_kernels):
    r"""Alias of `Compose('*', **kw_kernels)`.
    :math:`k_\otimes(X, Y) = \prod_{a \in \mathrm{features}} k_a(X_a, Y_a)`
    """
    return Compose('*', **kw_kernels)


def Additive(**kw_kernels):
    r"""Alias of `Compose('+', **kw_kernels)`.
    :math:`k_\oplus(X, Y) = \sum_{a \in \mathrm{features}} k_a(X_a, Y_a)`
    """
    return Compose('+', **kw_kernels)


def Convolution(kernel, normalize=True):
    r"""Creates a convolution kernel, which sums up evaluations of a base
    kernel on pairs of elements from two sequences.
    :math:`k_{CONV}(X, Y) = \sum_{x \in X} \sum_{y \in Y} k_{base}(x, y)`

    Parameters
    ----------
    kernel: base kernel
        The kernel can be any base kernel or a composition of base kernels in
        this module, while the attribute to be convolved should be
        fixed-length sequences.
    normalize: bool
        Whether or not to normalize the convolution to ensure it is a valid
        base kernel in range [0, 1].
    """

    @cpptype(kernel=kernel.dtype)
    class ConvolutionOf(BaseKernel):
        def __init__(self, kernel):
            self.kernel = kernel

        def __call__(self, X, Y, jac=False):
            if jac is True:
                Fxy, Jxy = list(zip(*[
                    self.kernel(x, y, jac=True) for x in X for y in Y
                ]))
                return np.sum(Fxy), np.sum(Jxy, axis=0)
            else:
                return np.sum([self.kernel(x, y) for x in X for y in Y])

        def __repr__(self):
            return f'Convolution({repr(self.kernel)})'

        def gen_expr(self, x, y, jac=False, theta_prefix=''):
            F, J = self.kernel.gen_expr(
                '_1', '_2', True, theta_prefix + 'kernel.'
            )
            f = Template(
                r'convolution([&](auto _1, auto _2){return ${f};}, ${x}, ${y})'
            ).render(
                x=x, y=y, f=F
            )
            if jac is True:
                template = Template(
                    r'''convolution_jacobian(
                            [&](auto _1, auto _2){return ${j};},
                            ${x},
                            ${y}
                        )'''
                )
                jacobian = [template.render(x=x, y=y, j=j) for j in J]
                return f, jacobian
            else:
                return f

        @property
        def theta(self):
            return namedtuple(
                'ConvolutionHyperparameters',
                ['kernel']
            )(self.kernel.theta)

        @theta.setter
        def theta(self, seq):
            self.kernel.theta = seq[0]

        @property
        def bounds(self):
            return (self.kernel.bounds,)

    if normalize is True:
        return Normalize(ConvolutionOf(kernel))
    else:
        return ConvolutionOf(kernel)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from collections import OrderedDict
import numpy as np
import scipy.spatial.distance
import sympy
from sympy.core.sympify import sympify
from sympy.utilities.autowrap import ufuncify


class RBFKernel:

    def __init__(self, expr, x, **hyperparameters):
        self.expr = sympify(expr)
        self._params = OrderedDict(**hyperparameters)
        self._param = (
            sympy.symbols(x),
            *sympy.symbols(','.join(self._params.keys()))
        )
        self._fun = ufuncify(self._param, self.expr)
        self._grad = [ufuncify(self._param, sympy.diff(self.expr, t))
                      for t in self._params]

    def get_params(self):
        return self._params

    @property
    def theta(self):
        return np.log(list(self._params.values()))

    @theta.setter
    def theta(self, args):
        for k, v in zip(self._params, np.exp(args)):
            self._params[k] = v

    def __call__(self, X, Y=None):
        if Y is None:
            d = scipy.spatial.distance.cdist(X, X)
        else:
            d = scipy.spatial.distance.cdist(X, Y)
        return self._fun(d, *self._params.values())

    def gradient(self, X):
        d = scipy.spatial.distance.cdist(X, X)
        return [g(d, *self._params.values()) for g in self._grad]

    def diag(self, X):
        return self._fun(np.zeros(len(X)), *self._params.values())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Adaptor for ASE's Atom objects"""
from itertools import product
import uuid
import numpy as np
from scipy.spatial import cKDTree
from graphdot.graph.adjacency.atomic import AtomicAdjacency
from graphdot.minipandas import DataFrame


def _from_ase(cls, atoms, use_charge=False, use_pbc=True,
              adjacency='default'):
    """Convert from ASE atoms to molecular graph

    Parameters
    ----------
    atoms: ASE Atoms object
        A molecule as represented by a collection of atoms in 3D space.
    usb_pbc: boolean or list of 3 booleans
        Whether to use the periodic boundary condition as specified in the
        atoms object to create edges between atoms.
    adjacency: 'default' or object
        A functor that implements the rule for making edges between atoms.

    Returns
    -------
    graphdot.Graph:
        a molecular graph where atoms become nodes while edges resemble
        short-range interatomic interactions.
    """
    if adjacency == 'default':
        adjacency = AtomicAdjacency()

    nodes = DataFrame({'!i': range(len(atoms))})
    nodes['element'] = atoms.get_atomic_numbers().astype(np.int8)
    if use_charge:
        nodes['charge'] = atoms.get_initial_charges().astype(np.float32)

    pbc = np.logical_and(atoms.pbc, use_pbc)
    images = [(atoms.cell.T * image).sum(axis=1) for image in product(
        *tuple([-1, 0, 1] if p else [0] for p in pbc))]
    x = atoms.get_positions()
    x_images = np.vstack([x + i for i in images])
    # prefer lookup over integer modulo
    j_images = list(range(len(atoms))) * len(images)

    cutoff = adjacency.cutoff(atoms.get_atomic_numbers())
    nl = cKDTree(x).sparse_distance_matrix(cKDTree(x_images), cutoff)

    edgedict = {}
    for (i, j), r in nl.items():
        j = j_images[j]
        if j > i:
            w = adjacency(atoms[i].number, atoms[j].number, r)
            if w > 0 and ((i, j) not in edgedict or edgedict[(i, j)][1] > r):
                edgedict[(i, j)] = (w, r)
    i, j, w, r = list(zip(*[(i, j, w, r)
                            for (i, j), (w, r) in edgedict.items()]))

    edges = DataFrame({
        '!i': np.array(i, dtype=np.uint32),
        '!j': np.array(j, dtype=np.uint32),
        '!w': np.array(w, dtype=np.float32),
        'length': np.array(r, dtype=np.float32),
    })

    return cls(nodes, edges, title='Molecule {formula} {id}'.format(
                formula=atoms.get_chemical_formula(), id=uuid.uuid4().hex))



def flatten(iterable):
    for item in iterable:
        if hasattr(item, '__iter__'):
            yield from flatten(item)
        else:
            yield item


def fold_like(flat, example):
    folded = []
    for item in example:
        if hasattr(item, '__iter__'):
            n = len(list(flatten(item)))
            folded.append(fold_like(flat[:n], item))
            flat = flat[n:]
        else:
            folded.append(flat[0])
            flat = flat[1:]
    return tuple(folded)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
from collections import OrderedDict


def add_classmethod(cls, override=False):
    def decorate(func):
        if override is not True and hasattr(cls, func.__name__):
            raise RuntimeError('Class %s already has a method named %s' % (
                repr(cls), func.__name__))
        clsm = classmethod(func)
        clsm.__doc__ = "Add-on classmethod of %s\n\n%s" % (cls, func.__doc__)
        setattr(cls, func.__name__, clsm)
        return clsm
    return decorate


class Timer:
    def __init__(self):
        self.reset()

    def reset(self):
        self.t = OrderedDict()
        self.dt = OrderedDict()

    def tic(self, tag):
        self.t[tag] = time.perf_counter()

    def toc(self, tag):
        self.dt[tag] = time.perf_counter() - self.t[tag]
        del self.t[tag]

    def report(self, unit='s'):
        if unit == 's':
            scale = 1.0
        elif unit == 'ms':
            scale = 1e3
        elif unit == 'us':
            scale = 1e6
        elif unit == 'ns':
            scale = 1e9
        else:
            raise ValueError('Unknown unit %s' % unit)
        for tag, dt in self.dt.items():
            print('%9.1f %s on %s' % (dt * scale, unit, tag))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Adaptor for PyMatgen's Molecule objects"""
import pymatgen.io


def _from_pymatgen(cls, molecule, use_pbc=True, adjacency='default'):
    """Convert from pymatgen molecule to molecular graph

    Parameters
    ----------
    molecule: pymatgen Molecule object
        A molecule as represented by a collection of atoms in 3D space.
    usb_pbc: boolean or list of 3 booleans
        Whether to use the periodic boundary condition as specified in the
        atoms object to create edges between atoms.
    adjacency: 'default' or object
        A functor that implements the rule for making edges between atoms.

    Returns
    -------
    graphdot.Graph:
        a molecular graph where atoms become nodes while edges resemble
        short-range interatomic interactions.
    """
    atoms = pymatgen.io.ase.AseAtomsAdaptor.get_atoms(molecule)
    return cls.from_ase(atoms, use_pbc, adjacency)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import ctypes
from pycuda.driver import managed_empty
from pycuda.driver import mem_attach_flags as ma_flags


def unified_memory(size, dtype):
    return managed_empty(size, dtype, 'C', ma_flags.GLOBAL)


def pad(n, align):
    return ((n + align - 1) // align) * align


class ArrayView(np.ndarray):
    @property
    def ptr(self):
        return ctypes.addressof(ctypes.c_char.from_buffer(self.base))


class BeltAllocator:

    def __init__(self, dtype, slab_size=16777216, allocator=unified_memory):
        self.dtype = dtype
        self.slab_size = slab_size
        self.allocator = allocator
        self.current = None
        self.held = []

        self._allocate_slab()

    def __call__(self, size, alignment=8):
        if size > self.slab_size // 8:
            # no need of belt allocator for large allocations
            # return self.allocator(size, self.dtype)
            raise ValueError('Size %d too large for belt allocator' % size)
        else:
            self.head = pad(self.head, alignment)
            if self.head + size > self.slab_size:
                self.held.append(self.current)
                self._allocate_slab()
                self.head = 0

            alloc = np.frombuffer(self.current[self.head:self.head + size],
                                  dtype=self.dtype).view(ArrayView)
            self.head += size
            return alloc

    def allocate(self, size, alignment=8):
        if size > self.slab_size // 8:
            # no need of belt allocator for large allocations
            # return self.allocator(size, self.dtype)
            raise ValueError('Size %d too large for belt allocator' % size)
        else:
            self.head = pad(self.head, alignment)
            if self.head + size > self.slab_size:
                self._allocate_slab()

            alloc = self.current[self.head:self.head + size]
            self.head += size
            return alloc

    def dummy(self, size, alignment=8):
        return None

    def _allocate_slab(self):
        print('Allocating slab')
        if self.current:
            self.held.append(self.current)
        buffer = self.allocator(self.slab_size, self.dtype)
        buffer[:] = 0
        self.current = memoryview(buffer)
        self.head = 0


# if __name__ == '__main__':

#     import pycuda.autoinit

#     alloc = BeltAllocator(np.float32)

#     A = alloc(9)
#     B = alloc(1)
#     A[0] = 42
#     B[0] = 88
#     print(type(A))
#     print(A.base)
#     print(A.ptr)
#     print(B.ptr)
#     print(alloc.current[0])
#     print(np.frombuffer(alloc.current[0:32], np.float32))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import scipy.linalg


def chol_solve(A, b):
    L = np.linalg.cholesky(A)
    x = scipy.linalg.solve_triangular(
            L,
            scipy.linalg.solve_triangular(
                L, b,
                lower=True,
                check_finite=False),
            trans='C',
            lower=True,
            check_finite=False
        )
    return x


class CholSolver:
    def __init__(self, A):
        self.L = np.linalg.cholesky(A)

    def __call__(self, b):
        return scipy.linalg.solve_triangular(
            self.L,
            scipy.linalg.solve_triangular(
                self.L, b,
                lower=True,
                check_finite=False),
            trans='C',
            lower=True,
            check_finite=False
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import warnings

warnings.warn(
    "The basekernel module has been relocated to graphdot.kernel.basekernel.\n"
    "Please update relevant import statements from\n"
    "'from graphdot.kernel.marginalized.basekernel import *'\n"
    "to\n"
    "'from graphdot.kernel.basekernel import *'"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import warnings
import numpy as np
from pycuda.driver import managed_empty
from pycuda.driver import mem_attach_flags as ma_flags


def managed_allocator(count, dtype):
    return managed_empty(count, dtype, 'C', ma_flags.GLOBAL)


def numpy_allocator(count, dtype):
    return np.empty(count, dtype)


class ResizableArray:
    """
    Python version of std::vector. To be used together with CUDA managed memory
    to reduce kernel launch delay.
    """

    def __init__(self, dtype, count=0, allocator='managed'):
        self.dtype = dtype
        self.allocator = self._choose_allocator(allocator)
        self._ptr = None
        self._active = None
        self._size = 0
        self._capacity = 0
        if count:
            self.resize(count)

    def _choose_allocator(self, allocator):
        if isinstance(allocator, str):
            if allocator == 'managed':
                return managed_allocator
            elif allocator == 'numpy':
                return numpy_allocator
            else:
                raise ValueError('Unknown allocator "%s"' % allocator)
        else:
            return allocator

    def append(self, value):
        if self._size == self._capacity:
            self.reserve(self._capacity * 2 + 1)
        self._ptr[self._size] = value
        self._size += 1
        self._update_active()

    def resize(self, count):
        count = int(count)
        if count > self._capacity:
            self.reserve(count)
        self._size = count
        self._update_active()

    def reserve(self, count):
        count = int(count)
        if count <= self._size:
            warnings.warn('Reserving no more than current size has no effect')
            return
        _new = self.allocator(count, self.dtype)
        if self._size > 0:  # copy data to new buffer
            _new[:self._size] = self._ptr[:self._size]
        self._ptr = _new
        self._capacity = count
        self._update_active()

    def clear(self):
        self._size = 0
        self._active = None

    def __len__(self):
        return self._size

    def __getitem__(self, key):
        return self._active[key]

    def __setitem__(self, key, value):
        self._active[key] = value

    def __iadd__(self, gen):
        lst = list(gen)
        ext_size = self._size + len(lst)
        self.reserve(ext_size)
        self._ptr[self._size:ext_size] = lst
        self._size = ext_size
        self._update_active()
        return self

    @property
    def data(self):
        return self._ptr

    @property
    def capacity(self):
        return self._capacity

    def _update_active(self):
        self._active = self._ptr[:self._size]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from pycuda.driver import (managed_empty, managed_empty_like, managed_zeros)
from pycuda.driver import mem_attach_flags as ma_flags


class ManagedArray(np.ndarray):
    @property
    def ptr(self):
        return int(self.base.base)


def umempty(size, dtype=np.float32):
    return managed_empty(size, dtype, 'C', ma_flags.GLOBAL)


def umzeros(size, dtype=np.float32):
    return managed_zeros(size, dtype, 'C', ma_flags.GLOBAL)


def umlike(array):
    u = managed_empty_like(array, ma_flags.GLOBAL)
    u[:] = array[:]
    return u


def umarray(array):
    u = managed_empty_like(array, ma_flags.GLOBAL)
    u[:] = array[:]
    return u



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from graphdot.kernel.marginalized import MarginalizedGraphKernel
from graphdot.kernel.basekernel import KroneckerDelta
from graphdot.kernel.basekernel import SquareExponential
from graphdot.kernel.basekernel import TensorProduct


class Tang2019MolecularKernel:
    """A margianlized graph kernel for **3D molecular structures** as in:
    Tang, Y. H., & de Jong, W. A. (2019). Prediction of atomization energy
    using graph kernel and active learning. *The Journal of chemical physics*,
    150(4), 044107.
    The kernel can be directly used together with Graph.from_ase() to operate
    on molecular structures.

    Parameters
    ----------
    stopping_probability: float in (0, 1)
        The probability for the random walk to stop during each step.
    starting_probability: float
        The probability for the random walk to start from any node. See the `p`
        kwarg of :class:`graphdot.kernel.marginalized.MarginalizedGraphKernel`
    element_prior: float in (0, 1)
        The baseline similarity between distinct elements --- an element
        always have a similarity 1 to itself.
    edge_length_scale: float in (0, inf)
        length scale of the Gaussian kernel on edge length. A rule of thumb is
        that the similarity decays smoothly from 1 to nearly 0 around three
        times of the length scale.
    """

    def __init__(self,
                 stopping_probability=0.01,
                 starting_probability='uniform',
                 element_prior=0.2,
                 edge_length_scale=0.05, **kwargs):
        self.stopping_probability = stopping_probability
        self.starting_probability = starting_probability
        self.element_prior = element_prior
        self.edge_length_scale = edge_length_scale
        self._makekernel(**kwargs)

    def _makekernel(self, **kwargs):
        self.kernel = MarginalizedGraphKernel(
            TensorProduct(element=KroneckerDelta(self.element_prior)),
            TensorProduct(length=SquareExponential(self.edge_length_scale)),
            q=self.stopping_probability,
            p=self.starting_probability,
            **kwargs
        )

    def __call__(self, X, Y=None, **kwargs):
        """Same call signature as
        :py:meth:`graphdot.kernel.marginalized.MarginalizedGraphKernel.__call__`
        """
        return self.kernel(X, Y, **kwargs)

    def diag(self, X, **kwargs):
        """Same call signature as
        :py:meth:`graphdot.kernel.marginalized.MarginalizedGraphKernel.diag`
        """
        return self.kernel.diag(X, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import scipy.sparse
import scipy.sparse.linalg
from graphdot.graph import Graph
from graphdot.graph.adjacency.atomic import AtomicAdjacency
from graphdot.kernel.basekernel import TensorProduct
from graphdot.kernel.basekernel import KroneckerDelta
from graphdot.kernel.basekernel import SquareExponential


class M3:
    """The Marginalized MiniMax (M3) metric between molecules"""

    def __init__(self, use_charge=False, adjacency='default', q=0.01,
                 element_delta=0.2, bond_eps=0.02, charge_eps=0.2):

        self.use_charge = use_charge
        if adjacency == 'default':
            self.adjacency = AtomicAdjacency(shape='tent2', zoom=0.75)
        else:
            self.adjacency = adjacency
        self.q = q
        if use_charge:
            self.node_kernel = TensorProduct(
                element=KroneckerDelta(element_delta),
                charge=SquareExponential(charge_eps),
            )
        else:
            self.node_kernel = TensorProduct(
                element=KroneckerDelta(element_delta)
            )
        self.edge_kernel = TensorProduct(length=SquareExponential(bond_eps))

    def __call__(self, atoms1, atoms2):
        args = dict(use_charge=self.use_charge, adjacency=self.adjacency)
        g1 = Graph.from_ase(atoms1, **args)
        g2 = Graph.from_ase(atoms2, **args)

        R1 = self._mlgk(g1, g1).diagonal()**-0.5
        R2 = self._mlgk(g2, g2).diagonal()**-0.5
        R12 = self._mlgk(g1, g2)

        K = R1[:, None] * R12 * R2[None, :]
        D = np.sqrt(np.maximum(2 - 2 * K, 0))

        return max(D.min(axis=1).max(), D.min(axis=0).max())

    def _mlgk(self, g1, g2):
        n1 = len(g1.nodes)
        n2 = len(g2.nodes)

        A1 = scipy.sparse.csc_matrix(
            (g1.edges['!w'], (g1.edges['!i'], g1.edges['!j'])),
            (n1, n1)
        )
        A2 = scipy.sparse.csc_matrix(
            (g2.edges['!w'], (g2.edges['!i'], g2.edges['!j'])),
            (n2, n2)
        )
        A1 += A1.T
        A2 += A2.T

        d1 = np.asarray(A1.sum(axis=0)).ravel()
        d2 = np.asarray(A2.sum(axis=0)).ravel()

        Ax = scipy.sparse.kron(A1, A2)

        Vx = np.array(
            [self.node_kernel(atom1, atom2)
             for atom1 in g1.nodes.itertuples()
             for atom2 in g2.nodes.itertuples()])

        E, i, j = [], [], []
        for i1, j1, e1 in zip(g1.edges['!i'],
                              g1.edges['!j'],
                              g1.edges.itertuples()):
            for i2, j2, e2 in zip(g2.edges['!i'],
                                  g2.edges['!j'],
                                  g2.edges.itertuples()):
                e = self.edge_kernel(e1, e2)
                E += [e, e, e, e]
                i += [i1 * n2 + i2,
                      j1 * n2 + i2,
                      j1 * n2 + j2,
                      i1 * n2 + j2]
                j += [j1 * n2 + j2,
                      i1 * n2 + j2,
                      i1 * n2 + i2,
                      j1 * n2 + i2]
        Ex = scipy.sparse.csc_matrix((E, (i, j)), (n1 * n2, n1 * n2))

        qx = np.kron(np.ones(n1), np.ones(n2))
        Dx = np.kron(d1, d2) / (1 - self.q)**2
        rhs = Dx * qx
        Y = scipy.sparse.diags([Dx / Vx], [0]) - Ax.multiply(Ex)
        R, _ = scipy.sparse.linalg.cg(
            Y, rhs,
            M=scipy.sparse.diags([Vx / Dx], [0]),
            atol=1e-7
        )

        return R.reshape(n1, n2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-import itertools as it
import numpy as np


_convertible = {
    # lvalue : rvalue
    'b': 'b',
    'i': 'iu',
    'u': 'iu',
    'f': 'f',
    'c': 'c',
    'm': 'm',
    'M': 'M',
    'O': 'O',
    'S': 'S',
    'U': 'U',
    'V': 'V'
}


def can_cast(src, dst):
    return src.kind in _convertible[dst.kind]


class common_min_type:
    @staticmethod
    def of_values(iterable, coerce=True, min_float=np.float32, signed=True):
        '''Find the common minimum elemet type that can safely hold all
        elements of an iterable sequence.

        Parameters
        ----------
        iterable:
            Sequence of objects for which a common type is to be inferred.
        coerce: bool
            Whether or not to up-cast in case when elements have different
            but inter-converible types.
        min_float: dtype
            The smallest floating-point type that should be returned.
            Introduced because float16 is not universally supported yet.

        Returns
        -------
        t: np.dtype or class
            If `coerce=True`, t would be the smallest numpy type that can
            safely contain all the values; otherwise, t would either be the
            smallest numpy dtype or the Python class that the elements
            belongs to.
        '''
        t = None
        for i in iterable:
            r = np.min_scalar_type(i) if np.isscalar(i) else type(i)
            if signed and isinstance(r, np.dtype) and r.kind == 'u':
                r = np.promote_types(r, np.int8)
            t = t or r
            if t != r:
                if coerce:
                    t = np.promote_types(t, r)
                else:
                    return None

        if isinstance(t, np.dtype) and t.kind == 'f':
            t = np.promote_types(t, min_float)

        return t

    @staticmethod
    def of_types(types, coerce=True, min_float=np.float32, signed=True):
        '''Find the common minimum elemet type that can safely hold all types
        in the given list.

        Parameters
        ----------
        iterable:
            Sequence of types for which a common type is to be inferred.
        coerce: bool
            Whether or not to up-cast in case when elements have different
            but inter-converible types.
        min_float: dtype
            The smallest floating-point type that should be returned.
            Introduced because float16 is not universally supported yet.

        Returns
        -------
        t: np.dtype or class
            If `coerce=True`, t would be the smallest numpy type that can
            safely contain all the values; otherwise, t would either be the
            smallest numpy dtype or the Python class that the elements
            belongs to.
        '''
        t = next(iter(types))
        for r in types:
            if signed and isinstance(r, np.dtype) and r.kind == 'u':
                r = np.promote_types(r, np.int8)
            if t != r:
                if coerce:
                    t = np.promote_types(t, r)
                else:
                    return None

        if isinstance(t, np.dtype) and t.kind == 'f':
            t = np.promote_types(t, min_float)

        return t


class common_concrete_type:
    @staticmethod
    def of_values(iterable):
        '''Find the common concrete type (if one exists) of all elements in a
        sequence.

        Returns
        -------
        t: type or None
            Returns the type of the elements if all elements share a common
            type, and returns `None` otherwise.
        '''
        return common_concrete_type.of_types(map(type, iterable))

    @staticmethod
    def of_types(types):
        '''Find the common concrete type (if one exists) among a type list.

        Returns
        -------
        t: type or None
            Returns the type of the elements if all elements share a common
            type, and returns `None` otherwise.
        '''
        t = next(iter(types))
        for i in types:
            if t != i:
                return None
        return t


def have_same_fields(t1, t2):
    if bool(t1.fields) != bool(t2.fields):
        return False
    if t1.fields and t2.fields:
        if len(t1.fields) != len(t2.fields):
            return False
        if set(t1.fields) != set(t2.fields):
            return False
        for f in t1.fields:
            if not have_same_fields(t1.fields[f][0], t2.fields[f][0]):
                return False
    return True


class _dtype_util:
    @staticmethod
    def is_object(t):
        return t.names is not None

    @staticmethod
    def is_array(t):
        return t.subdtype is not None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from sympy.codegen import ast
from sympy.printing.cxxcode import CXX11CodePrinter


class CUDACXX11CodePrinter(CXX11CodePrinter):
    _ns = ''

    def __init__(self, settings):
        super().__init__(settings)

    def __call__(self, expr, symbol_to_variable):
        self.symbol_to_variable = symbol_to_variable
        return self.doprint(expr)

    def _print_Symbol(self, expr):
        name = self.symbol_to_variable[super()._print_Symbol(expr)]
        if expr in self._settings['dereference']:
            return '(*{0})'.format(name)
        else:
            return name


cudacxxcode = CUDACXX11CodePrinter(
    dict(
        user_functions={
            'Pow': [
                # if exp is positive integer
                (lambda b, e: e.is_integer and int(e) >= 0,
                 lambda b, e: 'graphdot::ipow<%d>(%s)' % (int(e), b)),
                # if exp is negative integer
                (lambda b, e: e.is_integer and int(e) < 0,
                 lambda b, e: 'graphdot::ripow<%d>(%s)' % (-int(e), b)),
                # otherwise
                (lambda b, e: True, 'powf'),
            ]
        },
        type_aliases={
            ast.real: ast.float32,
            ast.integer: ast.int32
        }
    )
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Adaptor for NetworkX's Graph objects"""
import networkx as nx
from graphdot.minipandas import DataFrame


def _from_networkx(cls, graph, weight=None):
    """Convert from NetworkX ``Graph``

    Parameters
    ----------
    graph: a NetworkX ``Graph`` instance
        an undirected graph with homogeneous node and edge attributes, i.e.
        carrying same attributes.
    weight: str
        name of the attribute that encode edge weights

    Returns
    -------
    graphdot.graph.Graph
        the converted graph
    """

    nodes = list(graph.nodes)

    if not all(isinstance(x, int) for x in nodes) \
            or max(nodes) + 1 != len(nodes) or min(nodes) < 0:
        graph = nx.relabel.convert_node_labels_to_integers(graph)

    ''' extrac title '''
    title = graph.graph['title'] if 'title' in graph.graph.keys() else ''

    ''' convert node attributes '''
    node_attr = []
    for index, node in graph.nodes.items():
        if index == 0:
            node_attr = sorted(node.keys())
        elif node_attr != sorted(node.keys()):
            # raise TypeError(f'Node {index} '
            #                 f'attributes {node.keys()} '
            #                 f'inconsistent with {node_attr}')
            raise TypeError('Node {} attributes {} '
                            'inconsistent with {}'.format(
                                index,
                                node.keys(),
                                node_attr))

    node_df = DataFrame({'!i': range(len(graph.nodes))})
    for key in node_attr:
        node_df[key] = [node[key] for node in graph.nodes.values()]

    ''' convert edge attributes '''
    edge_attr = []
    for index, ((i, j), edge) in enumerate(graph.edges.items()):
        if index == 0:
            edge_attr = sorted(edge.keys())
        elif edge_attr != sorted(edge.keys()):
            # raise TypeError(f'Edge {(i, j)} '
            #                 f'attributes {edge.keys()} '
            #                 f'inconsistent with {edge_attr}')
            raise TypeError('Edge {} attributes {} '
                            'inconsistent with {}'.format(
                                (i, j),
                                edge.keys(),
                                edge_attr
                            ))

    edge_df = DataFrame()
    if len(graph.edges.keys()) == 0:
        raise RuntimeError(f'Graph {graph} has no edges.')
    edge_df['!i'], edge_df['!j'] = zip(*graph.edges.keys())
    if weight is not None:
        edge_df['!w'] = [edge[weight] for edge in graph.edges.values()]
    for key in edge_attr:
        if key != weight:
            edge_df[key] = [edge[key] for edge in graph.edges.values()]

    return cls(nodes=node_df, edges=edge_df, title=title)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" The GraphDot library for graph similarity comparison """
from .graph import Graph

__all__ = ['Graph']

__version__ = '0.5a1'
__author__ = 'Yu-Hang "Maxin" Tang, Oguz Selvitopi'
__maintainer__ = 'Yu-Hang "Maxin" Tang'
__email__ = 'Tang.Maxin@gmail.com'
__license__ = 'see LICENSE file'
__copyright__ = '''
GraphDot Copyright (c) 2019, The Regents of the University of California,
through Lawrence Berkeley National Laboratory (subject to receipt of any
required approvals from the U.S. Dept. of Energy).  All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit other to do
so.
'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""GraphDot's native graph container class

This module defines the class ``Graph`` that are used to store graphs across
this library, and provides conversion and importing methods from popular
graph formats.
"""
import itertools as it
import numpy as np
from graphdot.codegen.typetool import common_min_type
from graphdot.minipandas import DataFrame
from ._from_ase import _from_ase
from ._from_networkx import _from_networkx
from ._from_pymatgen import _from_pymatgen
from ._from_rdkit import _from_rdkit


__all__ = ['Graph']


def _from_dict(d):
    if isinstance(d, DataFrame):
        return d
    else:
        # dict of column name-data pairs
        return DataFrame(d)


class Graph:
    """
    This is the class that stores a graph in GraphDot.

    Parameters
    ----------
    nodes: dataframe
        each row represent a node
    edges: dataframe
        each row represent an edge
    title: str
        a unique identifier of the graph
    """

    def __init__(self, nodes, edges, title=''):
        self.title = str(title)
        self.nodes = _from_dict(nodes)
        self.edges = _from_dict(edges)

    def __repr__(self):
        return '<{}(nodes={}, edges={}, title={})>'.\
            format(type(self).__name__,
                   repr(self.nodes),
                   repr(self.edges),
                   repr(self.title))

    @staticmethod
    def has_unified_types(graphs):
        '''Check if all graphs have the same set of nodal/edge features.'''
        first = next(iter(graphs))
        node_t = first.nodes.rowtype()
        edge_t = first.edges.rowtype()
        for second in graphs:
            if second.nodes.rowtype() != node_t:
                return ('nodes', first, second)
            elif second.edges.rowtype() != edge_t:
                return ('edges', first, second)
        return True

    @classmethod
    def unify_datatype(cls, graphs, inplace=False):
        '''Ensure that each attribute has the same data type across graphs.

        Parameters
        ----------
        graphs: list
            A list of graphs that have the same set of node and edge
            features. The types for each attribute will then be
            chosen to be the smallest scalar type that can safely hold all the
            values as found across the graphs.
        inplace: bool
            Whether or not to modify the graph features in-place.

        Returns
        -------
        None or list
            If inplace is True, the graphs will be modified in-place and
            nothing will be returned. Otherwise, a new list of graphs with
            type-unified features will be returned.
        '''

        '''copy graphs if not editing in-place'''
        if inplace is not True:
            def shallowcopy(g):
                h = cls(
                    nodes=g.nodes.copy(deep=False),
                    edges=g.edges.copy(deep=False),
                    title=g.title
                )
                for key, val in g.__dict__.items():
                    if key not in ['nodes', 'edges', 'title']:
                        h.__dict__[key] = val
                return h
            graphs = [shallowcopy(g) for g in graphs]

        '''ensure all graphs have the same node and edge features'''
        features = {}
        for component in ['nodes', 'edges']:
            first = None
            for g in graphs:
                second = set(getattr(g, component).columns)
                first = first or second
                if second != first:
                    raise TypeError(
                        f'Graph {g} with node features {second} '
                        'does not match with the other graphs.'
                    )
            features[component] = first

        '''unify data type for each feature'''
        for component in ['nodes', 'edges']:
            group = [getattr(g, component) for g in graphs]
            for key in features[component]:
                types = [g[key].concrete_type for g in group]
                t = common_min_type.of_types(types)
                if t == np.object:
                    t = common_min_type.of_types(types, coerce=False)
                if t is None:
                    raise TypeError(
                        f'Cannot unify attribute {key} containing mixed '
                        'object types'
                    )

                if np.issctype(t):
                    for g in group:
                        g[key] = g[key].astype(t)
                elif t in [list, tuple, np.ndarray]:
                    t_sub = common_min_type.of_values(
                        it.chain.from_iterable(
                            it.chain.from_iterable([g[key] for g in group])
                        )
                    )
                    if t_sub is None:
                        raise TypeError(
                            f'Cannot find a common type for elements in {key}.'
                        )
                    for g in group:
                        g[key] = [np.array(seq, dtype=t_sub) for seq in g[key]]

        '''only returns if not editing in-place'''
        if inplace is not True:
            return graphs

    # @classmethod
    # def from_auto(cls, graph):
    #     # import importlib
    #     # graph_translator = {}
    #     #
    #     # if importlib.util.find_spec('ase') is not None:
    #     #     ase = importlib.import_module('ase')
    #     #
    #     #     def ase_translator(atoms):
    #     #         pass
    #     #
    #     #     graph_translator[ase.atomsatoms.Atoms] = ase_translator
    #     #
    #     # if importlib.util.find_spec('networkx') is not None:
    #     #     nx = importlib.import_module('networkx')
    #     #
    #     #     def networkx_graph_translator(atoms):
    #     #         pass
    #     #
    #     #     graph_translator[nx.Graph] = networkx_graph_translator
    #     pass

    @classmethod
    def from_networkx(cls, graph, weight=None):
        """Convert from NetworkX ``Graph``

        Parameters
        ----------
        graph: a NetworkX ``Graph`` instance
            an undirected graph with homogeneous node and edge features, i.e.
            carrying same features.
        weight: str
            name of the attribute that encode edge weights

        Returns
        -------
        graphdot.graph.Graph
            the converted graph
        """
        return _from_networkx(cls, graph, weight)

    @classmethod
    def from_ase(cls, atoms, use_charge=False, use_pbc=True,
                 adjacency='default'):
        """Convert from ASE atoms to molecular graph

        Parameters
        ----------
        atoms: ASE Atoms object
            A molecule as represented by a collection of atoms in 3D space.
        usb_pbc: boolean or list of 3 booleans
            Whether to use the periodic boundary condition as specified in the
            atoms object to create edges between atoms.
        adjacency: 'default' or object
            A functor that implements the rule for making edges between atoms.

        Returns
        -------
        graphdot.Graph:
            a molecular graph where atoms become nodes while edges resemble
            short-range interatomic interactions.
        """
        return _from_ase(cls, atoms, use_charge, use_pbc, adjacency)

    @classmethod
    def from_pymatgen(cls, molecule, use_pbc=True, adjacency='default'):
        """Convert from pymatgen molecule to molecular graph

        Parameters
        ----------
        molecule: pymatgen Molecule object
            A molecule as represented by a collection of atoms in 3D space.
        usb_pbc: boolean or list of 3 booleans
            Whether to use the periodic boundary condition as specified in the
            atoms object to create edges between atoms.
        adjacency: 'default' or object
            A functor that implements the rule for making edges between atoms.

        Returns
        -------
        graphdot.Graph:
            A molecular graph where atoms become nodes while edges resemble
            short-range interatomic interactions.
        """
        return _from_pymatgen(cls, molecule, use_pbc, adjacency)

    @classmethod
    def from_smiles(cls, smiles):
        """DEPRECATED and replaced by from_rdkit."""
        raise RuntimeError(
            'from_smiles has been removed, use from_rdkit instead.'
        )

    @classmethod
    def from_rdkit(cls, mol, bond_type='order', set_ring_list=True,
                   set_ring_stereo=True):
        """Convert a RDKit molecule to a graph

        Parameters
        ----------
        bond_type: 'order' or 'type'
            If 'order', an edge attribute 'order' will be populated with
            numeric values such as 1 for single bonds, 2 for double bonds, and
            1.5 for aromatic bonds. If 'type', an attribute 'type' will be
            populated with :py:class:`rdkit.Chem.BondType` values.
        set_ring_list: bool
            if True, a nodal attribute 'ring_list' will be used to store a list
            of the size of the rings that the atom participates in.
        set_ring_stereo: bool
            If True, an edge attribute 'ring_stereo' will be used to store the
            E-Z stereo configuration of substitutes at the end of each bond
            along a ring.

        Returns
        -------
        graphdot.Graph:
            A graph where nodes represent atoms and edges represent bonds. Each
            node and edge carries an array of features as inferred from the
            chemical structure of the molecule.
        """
        return _from_rdkit(cls, mol,
                           bond_type=bond_type,
                           set_ring_list=set_ring_list,
                           set_ring_stereo=set_ring_stereo)

    # @classmethod
    # def from_graphviz(cls, molecule):
    #     pass
    #
    # @classmethod
    # def from_dot(cls, molecule):
    #     """
    #     From the DOT graph description language
    #     https://en.wikipedia.org/wiki/DOT_(graph_description_language)
    #     """
    #     pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from collections import namedtuple
import numpy as np
import pandas as pd
from .series import Series


class DataFrame:
    def __init__(self, data=None):
        self._data = {}

        if isinstance(data, dict):
            for key, value in data.items():
                self[key] = value

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._data[key]
        elif hasattr(key, '__iter__'):
            return self.__class__({k: self._data[k] for k in key})
        else:
            raise TypeError(f'Invalid column index {key}')

    def __setitem__(self, key, value):
        self._data[key] = Series(value)

    def __getattr__(self, name):
        if name in self._data.keys():
            return self._data[name]
        else:
            raise AttributeError(f'Dataframe has no column {name}.')

    def __repr__(self):
        return repr(self._data)

    def __len__(self):
        return max([0] + [len(array) for array in self._data.values()])

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        for key in self._data:
            yield key

    @property
    def columns(self):
        return list(self._data.keys())

    def rowtype(self, pack=True):
        cols = np.array(list(self.columns))
        ctypes = {key: np.dtype(self[key].concrete_type) for key in cols}
        if pack is True:
            perm = np.argsort([-ctypes[key].itemsize for key in self.columns])
            cols = cols[perm]
        packed_dtype = np.dtype([(key, ctypes[key].newbyteorder('='))
                                for key in cols], align=True)
        return packed_dtype

    def rows(self):
        '''Iterate over rows in the form of named tuples while skipping columns
        that do not have valid field names.'''
        visible = [key for key in self._data if key.isidentifier()]

        class RowTuple(namedtuple('RowTuple', visible)):
            def __getitem__(self, key):
                '''To support both member access and index access.'''
                if isinstance(key, str):
                    return getattr(self, key)
                else:
                    return super().__getitem__(key)

        # for row in zip(*[self[key] for key in visible]):
        #     yield RowTuple(row)
        for i in range(len(self)):
            yield RowTuple(*[self[key][i] for key in visible])

    def itertuples(self):
        '''Alias of `rows()` for API compatibility with pandas.'''
        yield from self.rows()

    def iterrows(self):
        '''Iterate in (row_id, row_content) tuples.'''
        yield from enumerate(self.rows())

    def iterstates(self, pack=True):
        '''Iterate over rows, use the .state attribute if element is not
        scalar.'''
        cols = np.array(list(self.rowtype(pack=pack).fields.keys()))

        for row in zip(*[self[key] for key in cols]):
            yield tuple(i if np.isscalar(i) else i.state for i in row)

    def to_pandas(self):
        return pd.DataFrame(self._data)

    def copy(self, deep=False):
        if deep:
            return self.__class__({
                key: np.copy(value) for key, value in self._data.items()
            })
        else:
            return self.__class__(self._data)

    def drop(self, keys, inplace=False):
        if inplace is True:
            for key in keys:
                del self._data[key]
        else:
            return self[[k for k in self.columns if k not in keys]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import itertools as it
import numpy as np
from graphdot.codegen import Template
from .typetool import can_cast, _dtype_util


def cpptype(decls=[], **kwdecls):
    """
    cpptype is a class decorator that simplifies the translation of python
    objects to corresponding C++ structures.
    """
    ctype = np.dtype(decls + list(kwdecls.items()), align=True)

    def decor(cls):
        class CppType(type):
            @property
            def dtype(self):
                '''Enables numpy.dtype(cls) to return the layout of the C++
                counterpart.'''
                return ctype

            def __repr__(self):
                return f'@cpptype({repr(ctype)})\n{repr(cls)}'

        class Class(cls, metaclass=CppType):

            @property
            def dtype(self):
                '''Useful in base kernel composition.'''
                return Class.dtype

            @property
            def state(self):
                state = []
                for key, (field, _) in Class.dtype.fields.items():
                    if _dtype_util.is_object(field):
                        state.append(getattr(self, key).state)
                    elif _dtype_util.is_array(field):
                        states = getattr(self, key)
                        if not isinstance(states, np.ndarray):
                            raise TypeError(
                                f'attribute {key} declared as an array'
                                f'but the actual type is not a numpy'
                                f'ndarray.'
                            )
                        if states.shape != field.shape:
                            raise ValueError(
                                f'attribute {key}: '
                                f'actual shape of {states.shape} does not '
                                f'match declared shape of {field.shape}.'
                            )
                        if _dtype_util.is_object(field.base):
                            states = [
                                states[coord].state for coord
                                in it.product(*map(range, field.shape))
                            ]
                        state.append(
                            np.array(states, dtype=field.base)
                              .reshape(field.shape).tolist())
                    else:
                        state.append(field.type(getattr(self, key)))
                return tuple(state)

            def __setattr__(self, name, value):
                if name in Class.dtype.names:
                    lt = Class.dtype.fields[name][0]
                    if _dtype_util.is_array(lt):
                        if not isinstance(value, np.ndarray):
                            value = np.array(value)
                        if value.shape != lt.shape:
                            raise ValueError(
                                f"Cannot set array attribute '{name}' with "
                                f"value of mismatching shape:\n"
                                f"{value}"
                            )
                        for t in map(np.dtype, value.ravel()):
                            if not can_cast(t, lt.base):
                                raise TypeError(
                                    f"Cannot set attribute '{name}' "
                                    f"(C++ type {decltype(lt)}) "
                                    f"with values of {value.dtype}"
                                )
                    else:
                        rt = np.dtype(type(value))
                        if not can_cast(rt, lt):
                            raise TypeError(
                                f"Cannot set attribute '{name}' "
                                f"(C++ type {decltype(lt)}) "
                                f"with value {value} of {type(value)}"
                            )
                super().__setattr__(name, value)

            # TODO: need a nicer __repr__

        Class.__name__ = cls.__name__

        return Class

    return decor


def _assert_is_identifier(name):
    if name and not name.isidentifier():
        raise ValueError(
            f'Name {name} is not a valid Python/C++ identifier.'
        )


def decltype(t, name=''):
    '''Generate C++ source code for structures corresponding to a `cpptype`
    class.'''
    t = np.dtype(t, align=True)  # convert np.float32 etc. to dtype
    if name.startswith('$'):
        ''' template class types '''
        n, t, *Ts = name[1:].split('::')
        _assert_is_identifier(n)
        return Template(r'${template}<${arguments,}>${name}').render(
            template=t,
            arguments=[decltype(T) for T in Ts],
            name=n
        )
    else:
        _assert_is_identifier(name)
        if _dtype_util.is_object(t):
            ''' structs '''
            if len(t.names):
                return Template(r'struct{${members;};}${name}').render(
                    name=name,
                    members=[decltype(t.fields[v][0], v) for v in t.names])
            else:
                return f'constexpr static _empty {name} {{}}'
        elif _dtype_util.is_array(t):
            ''' C-style arrays '''
            return Template(r'${t} ${name}[${shape][}]').render(
                t=decltype(t.base),
                name=name,
                shape=t.shape
            )
        else:
            ''' scalar types and C-strings '''
            if t.kind == 'S':
                return f'char {name}[{t.itemsize}]'
            else:
                return f'{str(t.name)} {name}'.strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ctypes
import itertools as it
import numpy as np
from graphdot.codegen.cpptool import cpptype
from graphdot.codegen.typetool import common_min_type
from graphdot.cuda.array import umzeros, umempty, umlike

@cpptype(n_node=np.int32, n_octile=np.int32, p_degree=np.uintp,
         p_node=np.uintp, p_octile=np.uintp)
class OctileGraph:
    """ Python counterpart of C++ class graphdot::graph_t """

    @cpptype(elements=np.uintp, nzmask='<u8', nzmask_r='<u8',
             upper=np.int32, left=np.int32)
    class Octile:
        def __init__(self, upper, left, nzmask, nzmask_r, elements):
            self.upper = upper
            self.left = left
            self.nzmask = nzmask
            self.nzmask_r = nzmask_r
            self.elements = elements

    class CustomType:
        @cpptype(ptr=np.intp, size=np.int32)
        class FrozenArray(np.ndarray):
            @property
            def ptr(self):
                return ctypes.addressof(ctypes.c_char.from_buffer(self.base))

            @property
            def size(self):
                return len(self)

    def __init__(self, graph):

        self.nodes = nodes = graph.nodes.copy(deep=False)
        self.edges = edges = graph.edges.copy(deep=False)
        self.n_node = len(nodes)
        nnz = len(edges)

        ''' substitute columns corresponding to object-type node/edge
        attributes to their GPU counterparts '''
        for df in [nodes, edges]:
            for key in list(df.columns):
                if not np.issctype(df[key].dtype):
                    if issubclass(df[key].concrete_type, (list,
                                                          tuple,
                                                          np.ndarray)):
                        inner_type = common_min_type.of_types(
                            [x.dtype if isinstance(x, np.ndarray)
                             else common_min_type.of_values(x)
                             for x in df[key]]
                        )
                        if not np.issctype(inner_type):
                            raise(TypeError(
                                f'Expect scalar elements in tuple or list'
                                f'atttributes, got {inner_type}.'
                            ))
                        if not np.issctype(inner_type):
                            raise TypeError(
                                f'List-like graphs attribute must have scalar'
                                f'elements. Attribute {key} is {inner_type}.'
                            )
                        buffer = memoryview(umlike(np.fromiter(
                            it.chain.from_iterable(df[key]),
                            dtype=inner_type
                        )))
                        size = np.fromiter(map(len, df[key]), dtype=np.int)
                        head = np.cumsum(size) - size
                        # mangle key with type information
                        tag = '${key}::frozen_array::{dtype}'.format(
                            key=key,
                            dtype=inner_type.str
                        )
                        data = np.empty_like(df[key], dtype=np.object)
                        for i, (h, s) in enumerate(zip(head, size)):
                            data[i] = np.frombuffer(
                                buffer[h:h + s], dtype=inner_type
                            ).view(self.CustomType.FrozenArray)
                        df[tag] = data
                        df.drop([key], inplace=True)
                    else:
                        raise TypeError(
                            f'Unsupported non-scalar attribute {key} '
                            f'of type {df[key].concrete_type}'
                        )

        ''' add phantom label if none exists to facilitate C++ interop '''
        assert(len(nodes.columns) >= 1)
        if len(nodes.columns) == 1:
            nodes['labeled'] = np.zeros(len(nodes), np.bool_)

        assert(len(edges.columns) >= 2)
        if len(edges.columns) == 2:
            assert('!i' in edges and '!j' in edges)
            edges['labeled'] = np.zeros(len(edges), np.bool_)

        ''' determine node type '''
        nodes.drop(['!i'], inplace=True)
        self.node_t = node_t = nodes.rowtype()
        self.nodes_aos = umempty(len(nodes), dtype=node_t)
        self.nodes_aos[:] = list(nodes.iterstates())

        ''' determine whether graph is weighted, determine edge type,
            and compute node degrees '''
        self.degree = degree = umzeros(self.n_node, dtype=np.float32)
        edge_t = edges.drop(['!i', '!j', '!w']).rowtype()
        if '!w' in edges:  # weighted graph
            self.weighted = True
            np.add.at(degree, edges['!i'], edges['!w'])
            np.add.at(degree, edges['!j'], edges['!w'])

            if edge_t.itemsize != 0:
                labels = list(edges[edge_t.names].iterstates())
            else:
                labels = [None] * len(edges)

            edge_t = np.dtype(
                [('weight', np.float32), ('label', edge_t)],
                align=True
            )

            edges_aos = np.fromiter(zip(edges['!w'], labels), dtype=edge_t,
                                    count=nnz)
        else:
            self.weighted = False
            np.add.at(degree, edges['!i'], 1.0)
            np.add.at(degree, edges['!j'], 1.0)
            edges_aos = np.fromiter(edges[edge_t.names].iterstates(),
                                    dtype=edge_t, count=nnz)
        self.edge_t = edge_t
        degree[degree == 0] = 1.0

        ''' collect non-zero edge octiles '''
        indices = np.empty((4, nnz * 2), dtype=np.uint32, order='C')
        i, j, up, lf = indices
        i[:nnz] = edges['!i']
        j[:nnz] = edges['!j']
        # replicate & swap i and j for the lower triangular part
        indices[[0, 1], nnz:] = indices[[1, 0], :nnz]
        # get upper left corner of owner octiles
        up[:] = i - i % 8
        lf[:] = j - j % 8

        perm = np.lexsort(indices, axis=0)
        # Do not delete this line even though it looks useless. It does change
        # i, j, up, lf
        indices[:, :] = indices[:, perm]
        self.edges_aos = umempty(nnz * 2, edge_t)
        self.edges_aos[:] = edges_aos[perm % nnz]  # mod nnz due to symmetry

        diff = np.empty(nnz * 2)
        diff[1:] = (up[:-1] != up[1:]) | (lf[:-1] != lf[1:])
        diff[:1] = True
        oct_offset = np.flatnonzero(diff)
        self.n_octile = len(oct_offset)

        nzmasks = np.bitwise_or.reduceat(
            1 << (i - up + (j - lf) * 8).astype(np.uint64), oct_offset)
        nzmasks_r = np.bitwise_or.reduceat(
            1 << (j - lf + (i - up) * 8).astype(np.uint64), oct_offset)

        self.octiles = octiles = umempty(self.n_octile, self.Octile.dtype)
        octiles[:] = list(
            zip(int(self.edges_aos.base) + oct_offset * edge_t.itemsize,
                nzmasks,
                nzmasks_r,
                up[oct_offset],
                lf[oct_offset])
        )

    @property
    def p_octile(self):
        return int(self.octiles.base)

    @property
    def p_degree(self):
        return int(self.degree.base)

    @property
    def p_node(self):
        return int(self.nodes_aos.base)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Adaptor for RDKit's Molecule objects"""
import re
import networkx as nx
import numpy as np
from treelib import Tree
from rdkit.Chem import AllChem as Chem
from ._from_networkx import _from_networkx


class FunctionalGroup:
    """Functional Group.

    atom0 -> atom1 define a directed bond in the molecule. Then the bond is
    removed and the functional group is defined as a multitree. atom1 is the
    root node of the tree.

    Parameters
    ----------
    mol : molecule object in RDKit

    atom0, atom1 : atom object in RDKit

    depth: the depth of the multitree.

    Attributes
    ----------
    tree : multitree represent the functional group
        each node has 3 important attributes: tag: [atomic number, bond order
        with its parent], identifier: atom index defined in RDKit molecule
        object, data: RDKit atom object.

    """

    def __init__(self, mol, atom0, atom1, depth=5):
        self.mol = mol
        tree = Tree()
        bond_order = mol.GetBondBetweenAtoms(
            atom0.GetIdx(),
            atom1.GetIdx()
        ).GetBondTypeAsDouble()
        tree.create_node(
            tag=[atom0.GetAtomicNum(), bond_order],
            identifier=atom0.GetIdx(), data=atom0
        )
        tree.create_node(
            tag=[atom1.GetAtomicNum(), bond_order],
            identifier=atom1.GetIdx(),
            data=atom1,
            parent=atom0.GetIdx()
        )
        for _ in range(depth):
            for node in tree.all_nodes():
                if node.is_leaf():
                    for atom in node.data.GetNeighbors():
                        tree_id = tree._identifier
                        if atom.GetIdx() != node.predecessor(tree_id=tree_id):
                            order = mol.GetBondBetweenAtoms(
                                atom.GetIdx(),
                                node.data.GetIdx()
                            ).GetBondTypeAsDouble()
                            identifier = atom.GetIdx()
                            while tree.get_node(identifier) is not None:
                                identifier += len(mol.GetAtoms())
                            tree.create_node(
                                tag=[atom.GetAtomicNum(), order],
                                identifier=identifier,
                                data=atom,
                                parent=node.identifier
                            )
        self.tree = tree

    def __eq__(self, other):
        if self.get_rank_list() == other.get_rank_list():
            return True
        else:
            return False

    def __lt__(self, other):
        if self.get_rank_list() < other.get_rank_list():
            return True
        else:
            return False

    def __gt__(self, other):
        if self.get_rank_list() > other.get_rank_list():
            return True
        else:
            return False

    def get_rank_list(self):
        rank_list = []
        expand_tree = self.tree.expand_tree(mode=Tree.WIDTH, reverse=True)
        for identifier in expand_tree:
            rank_list += self.tree.get_node(identifier).tag
        return rank_list


def get_bond_orientation_dict(mol):
    bond_orientation_dict = {}
    mb = Chem.MolToMolBlock(mol, includeStereo=True, kekulize=False)
    for i, j, _, d in re.findall(r'^[\s*(\d+)]{4}$', mb, re.MULTILINE):
        i, j, d = int(i) - 1, int(j) - 1, int(d)
        i, j = min(i, j), max(i, j)
        bond_orientation_dict[(i, j)] = d
    return bond_orientation_dict


def get_atom_ring_stereo(mol, atom, ring_idx, depth=5,
                         bond_orientation_dict=None):
    """Return an atom is upward or downward refer to a ring plane.

    For atom in a ring. If it has 4 bonds. Two of them are included in the
    ring. Other two connecting 2 functional groups, has opposite orientation
    reference to the ring plane. Assuming the ring is in a plane, then the 2
    functional groups are assigned as upward and downward.

    Parameters
    ----------
    mol : molecule object in RDKit

    atom : atom object in RDKit

    ring_idx : a tuple of all index of atoms in the ring

    depth : the depth of the functional group tree

    bond_orientation_dict : a dictionary contains the all bond orientation
        information in the molecule

    Returns
    -------
    0 : No ring stereo.

    1 : The upward functional group is larger

    -1 : The downward functional group is larger

    """
    if bond_orientation_dict is None:
        bond_orientation_dict = get_bond_orientation_dict(mol)

    up_atom = down_atom = None
    updown_tag = None
    # bond to 2 hydrogen
    if len(atom.GetNeighbors()) == 2:
        return 0
    if len(atom.GetNeighbors()) > 4:
        raise RuntimeError(
            'cannot deal with atom in a ring with more than 4 bonds'
        )
    for bond in atom.GetBonds():
        # for carbon atom, atom ring stereo may exist if it has 4 single bonds.
        if bond.GetBondType() != Chem.BondType.SINGLE \
                and atom.GetAtomicNum() == 6:
            return 0
        i = bond.GetBeginAtom().GetIdx()
        j = bond.GetEndAtom().GetIdx()
        ij = (i, j)
        # skip bonded atoms in the ring
        if i in ring_idx and j in ring_idx:
            # in RDKit, the orientation information may saved in ring bond for
            # multi-ring molecules. The information is saved.
            if bond_orientation_dict.get(ij) != 0:
                updown_tag = bond_orientation_dict.get(ij)
            continue
        # get upward atom
        if bond_orientation_dict.get(ij) == 1:
            if up_atom is not None:
                raise Exception('2 bond orient up')
            temp = list(ij)
            temp.remove(atom.GetIdx())
            up_atomidx = temp[0]
            up_atom = mol.GetAtomWithIdx(up_atomidx)
        # get downward atom
        elif bond_orientation_dict.get(ij) == 6:
            if down_atom is not None:
                raise Exception('2 bond orient down')
            temp = list(ij)
            temp.remove(atom.GetIdx())
            down_atomidx = temp[0]
            down_atom = mol.GetAtomWithIdx(down_atomidx)
    # maybe there is bug for complex molecule
    if up_atom is None and down_atom is None:
        if updown_tag == 1:
            return 1
        elif updown_tag == 6:
            return -1
        else:
            return 0
    elif up_atom is None:
        return -1
    elif down_atom is None:
        return 1
    else:
        fg_up = FunctionalGroup(mol, atom, up_atom, depth)
        fg_down = FunctionalGroup(mol, atom, down_atom, depth)
        if fg_up > fg_down:
            return 1
        elif fg_up < fg_down:
            return -1
        else:
            return 0


def get_ringlist(mol):
    ringlist = [[] for _ in range(mol.GetNumAtoms())]
    for ring in mol.GetRingInfo().AtomRings():
        for i in ring:
            ringlist[i].append(len(ring))
    return [sorted(rings) if len(rings) else [0] for rings in ringlist]


def _from_rdkit(cls, mol, bond_type='order', set_ring_list=True,
                set_ring_stereo=True):
    g = nx.Graph()

    for i, atom in enumerate(mol.GetAtoms()):
        g.add_node(i)
        g.nodes[i]['atomic_number'] = atom.GetAtomicNum()
        g.nodes[i]['charge'] = atom.GetFormalCharge()
        g.nodes[i]['hcount'] = atom.GetTotalNumHs()
        g.nodes[i]['hybridization'] = atom.GetHybridization()
        g.nodes[i]['aromatic'] = atom.GetIsAromatic()
        g.nodes[i]['chiral'] = 0 if atom.IsInRing() else atom.GetChiralTag()

    if set_ring_list:
        for i, rings in enumerate(get_ringlist(mol)):
            g.nodes[i]['ring_list'] = rings

    for bond in mol.GetBonds():
        ij = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
        g.add_edge(*ij)
        if bond_type == 'order':
            g.edges[ij]['order'] = bond.GetBondTypeAsDouble()
        else:
            g.edges[ij]['type'] = bond.GetBondType()
        g.edges[ij]['aromatic'] = bond.GetIsAromatic()
        g.edges[ij]['conjugated'] = bond.GetIsConjugated()
        g.edges[ij]['stereo'] = bond.GetStereo()
        if set_ring_stereo is True:
            g.edges[ij]['ring_stereo'] = 0

    if set_ring_stereo is True:
        bond_orientation_dict = get_bond_orientation_dict(mol)
        for ring_idx in mol.GetRingInfo().AtomRings():
            atom_updown = []
            for idx in ring_idx:
                atom = mol.GetAtomWithIdx(idx)
                atom_updown.append(
                    get_atom_ring_stereo(
                        mol,
                        atom,
                        ring_idx,
                        depth=5,
                        bond_orientation_dict=bond_orientation_dict
                    )
                )
            atom_updown = np.array(atom_updown)
            non_zero_index = np.where(atom_updown != 0)[0]
            for j in range(len(non_zero_index)):
                b = non_zero_index[j]
                if j == len(non_zero_index) - 1:
                    e = non_zero_index[0]
                    length = len(atom_updown) + e - b
                else:
                    e = non_zero_index[j + 1]
                    length = e - b
                StereoOfRingBond = atom_updown[b] * atom_updown[e] / length
                for k in range(length):
                    idx1 = b + k if b + k < len(ring_idx)\
                        else b + k - len(ring_idx)
                    idx2 = b + k + 1 if b + k + 1 < len(ring_idx)\
                        else b + k + 1 - len(ring_idx)
                    ij = (ring_idx[idx1], ring_idx[idx2])
                    ij = (min(ij), max(ij))
                    g.edges[ij]['ring_stereo'] = StereoOfRingBond

    return _from_networkx(cls, g)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import copy
import uuid
import warnings
import functools
import numpy as np
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
import graphdot.cuda
from graphdot import cpp
from graphdot.codegen import Template
from graphdot.codegen.cpptool import decltype
from graphdot.cuda.array import umempty, umzeros, umarray, umlike
from graphdot.cuda.resizable_array import ResizableArray
from graphdot.kernel.basekernel import TensorProduct, _Multiply
from ._backend import Backend
from ._scratch import BlockScratch
from ._octilegraph import OctileGraph


class CUDABackend(Backend):
    """

    Parameters
    ----------
    context: :py:class:`pycuda.driver.Context` instance
        The CUDA context for launching kernels, Will use a default one if
        none is given.
    block_per_sm: int
        Tunes the GPU kernel.
    block_size: int
        Tunes the GPU kernel.
    """

    @staticmethod
    def array(ndarray):
        return umarray(ndarray)

    @staticmethod
    def zeros(size, dtype=np.float32):
        return umzeros(size, dtype)

    @staticmethod
    def empty(size, dtype=np.float32):
        return umempty(size, dtype)

    def __init__(self, **kwargs):
        self.ctx = kwargs.pop('cuda_context', graphdot.cuda.defctx)
        self.device = self.ctx.get_device()
        self.scratch = None
        self.scratch_capacity = 0
        self.graph_cache = {}
        self.graph_cpp = ResizableArray(OctileGraph.dtype, allocator='numpy')

        self.block_per_sm = kwargs.pop('block_per_sm', 8)
        self.block_size = kwargs.pop('block_size', 128)

        self.nvcc_extra = kwargs.pop('nvcc_extra', [])
        self._source = ''
        self._module = None

    def __deepcopy__(self, memo):
        return copy.copy(self)

    def _assert_homogeneous(self, x, y):
        try:
            assert(x.weighted == y.weighted)
            assert(x.node_t == y.node_t)
            assert(x.edge_t == y.edge_t)
        except AssertionError as e:
            raise TypeError(
                f'All nodes/edges must be of the same type: {str(e)}'
                'If the graph attributes match in name but differ in type, '
                'try to normalize automatically with `Graph.normalize_types`.'
            )

    def _allocate_scratch(self, count, capacity):
        if (self.scratch is None or len(self.scratch) < count or
                self.scratch[0].capacity < capacity):
            self.ctx.synchronize()
            self.scratch = [BlockScratch(capacity) for _ in range(count)]
            self.scratch_d = umarray(np.array([s.state for s in self.scratch],
                                              BlockScratch.dtype))
            self.scratch_capacity = self.scratch[0].capacity
            self.ctx.synchronize()

    def _register_graph(self, graph, p):
        if not hasattr(graph, 'uuid'):
            graph.uuid = uuid.uuid4()
        if graph.uuid not in self.graph_cache:
            # convert to GPU format
            og = OctileGraph(graph)
            # assign starting probabilities
            ps = umarray(np.array([p(*r) for r in graph.nodes.iterrows()],
                                  dtype=np.float32))
            i = len(self.graph_cpp)
            self.graph_cache[graph.uuid] = (i, ps, og)
            self.graph_cpp.append(og.state)
        return self.graph_cache[graph.uuid]

    def _compile(self, src):
        with warnings.catch_warnings(record=True) as w:
            module = SourceModule(
                src,
                options=['-std=c++14',
                         '-O4',
                         '--use_fast_math',
                         '--expt-relaxed-constexpr',
                         '--maxrregcount=64',
                         '-Xptxas', '-v',
                         '-lineinfo',
                         ] + self.nvcc_extra,
                no_extern_c=True,
                include_dirs=cpp.__path__)
        return module, [str(rec.message) for rec in w]

    @property
    @functools.lru_cache(maxsize=1)
    def template(self):
        return Template(os.path.join(os.path.dirname(__file__), 'template.cu'))

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if self.source != source:
            self._source = source
            self._module = None

    @property
    def module(self):
        if not self._module:
            self._module, self._compiler_message = self._compile(self.source)
        return self._module

    @staticmethod
    def gencode_kernel(kernel, name):
        fun, jac = kernel.gen_expr('x1', 'x2', jac=True)

        kernel_src = Template(r'''
        using ${name}_theta_t = ${theta_t};

        struct ${name}_t : ${name}_theta_t {

            constexpr static int jac_dims = ${jac_dims};

            template<class X> __device__ __inline__
            auto operator() (X const &x1, X const &x2) const {
                return ${expr};
            }

            template<class X> __device__ __inline__
            void _j_a_c_o_b_i_a_n_(float j[], X const &x1, X const &x2) const {
                ${jac;\n};
            }
        };

        __constant__ ${name}_t ${name};
        ''').render(
            name=name,
            jac_dims=len(jac),
            theta_t=decltype(kernel),
            expr=fun,
            jac=[f'j[{i}] = {expr}' for i, expr in enumerate(jac)],
        )

        return kernel_src

    def __call__(self, graphs, node_kernel, edge_kernel, p, q, jobs, starts,
                 output, output_shape, traits, timer):
        ''' transfer graphs and starting probabilities to GPU '''
        timer.tic('transferring graphs to GPU')

        og_last = None
        og_indices = np.empty(len(graphs), np.uint32)
        starting_p = umempty(len(graphs), np.uintp)

        for i, g in enumerate(graphs):
            idx, ps, og = self._register_graph(g, p)
            if i > 0:
                self._assert_homogeneous(og_last, og)
            og_last = og
            og_indices[i] = idx
            starting_p[i] = int(ps.base)

        og_indices_d = umlike(self.graph_cpp[og_indices])

        weighted = og_last.weighted
        node_t = og_last.node_t
        edge_t = og_last.edge_t

        timer.toc('transferring graphs to GPU')

        ''' allocate global job counter '''
        timer.tic('allocate global job counter')
        i_job_global = umzeros(1, np.uint32)
        timer.toc('allocate global job counter')

        ''' code generation '''
        timer.tic('code generation')
        if weighted:
            edge_kernel = TensorProduct(weight=_Multiply(),
                                        label=edge_kernel)

        node_kernel_src = self.gencode_kernel(node_kernel, 'node_kernel')
        edge_kernel_src = self.gencode_kernel(edge_kernel, 'edge_kernel')

        with self.template.context(traits=traits) as template:
            self.source = template.render(
                node_kernel=node_kernel_src,
                edge_kernel=edge_kernel_src,
                node_t=decltype(node_t),
                edge_t=decltype(edge_t)
            )
        timer.toc('code generation')

        ''' JIT '''
        timer.tic('JIT')
        kernel = self.module.get_function('graph_kernel_solver')
        timer.toc('JIT')

        ''' calculate launch configuration '''
        timer.tic('calculating launch configuration')
        launch_block_count = (self.device.MULTIPROCESSOR_COUNT
                              * self.block_per_sm)
        shmem_bytes_per_warp = self.module.get_global('shmem_bytes_per_warp')[1]
        shmem_bytes_per_block = (shmem_bytes_per_warp * self.block_size
                                 // self.device.WARP_SIZE)

        max_graph_size = np.max([len(g.nodes) for g in graphs])
        self._allocate_scratch(launch_block_count, 2 * max_graph_size**2)

        p_node_kernel, _ = self.module.get_global('node_kernel')
        cuda.memcpy_htod(p_node_kernel, np.array([node_kernel.state],
                                                 dtype=node_kernel.dtype))

        p_edge_kernel, _ = self.module.get_global('edge_kernel')
        cuda.memcpy_htod(p_edge_kernel, np.array([edge_kernel.state],
                                                 dtype=edge_kernel.dtype))

        timer.toc('calculating launch configuration')

        ''' GPU kernel execution '''
        timer.tic('GPU kernel execution')
        kernel(
            og_indices_d,
            starting_p,
            self.scratch_d,
            jobs,
            starts,
            output,
            i_job_global,
            np.uint32(len(jobs)),
            np.uint32(output_shape[0]),
            np.uint32(output_shape[1]),
            np.float32(q),
            np.float32(q),  # placeholder for q0
            grid=(launch_block_count, 1, 1),
            block=(self.block_size, 1, 1),
            shared=shmem_bytes_per_block,
        )
        self.ctx.synchronize()
        timer.toc('GPU kernel execution')



import io
import sys
import re
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

with open('graphdot/__init__.py') as fd:
    __version__ = re.search("__version__ = '(.*)'", fd.read()).group(1)

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read('README.md')


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox  # import here, cause outside the eggs aren't loaded
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


setup(
    name='graphdot',
    version=__version__,
    python_requires='>=3.6',
    url='https://gitlab.com/yhtang/graphdot',
    license='BSD',
    author='Yu-Hang Tang',
    tests_require=['tox'],
    install_requires=[
        'numpy>=1.16',
        'scipy>=1.3.0',
        'sympy>=1.3',
        'pandas>=0.24',
        'networkx>=2.4',
        'pycuda>=2019',
    ] + [
        'ase>=3.17',
        'pymatgen>=2019',
        'mendeleev'
    ],
    extras_require={
        'docs': ['sphinx', 'sphinx-rtd-theme'],
    },
    cmdclass={'test': Tox},
    author_email='Tang.Maxin@gmail.com',
    description='GPU-accelerated graph similarity algorithm library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude='test'),
    package_data={
        'graphdot': ['*/*.h', '*/*/*.h', '*/*/*.cu'],
    },
    # include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: C++',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
from collections import namedtuple
import numpy as np
from graphdot.graph import Graph
from graphdot.util import Timer
from ._backend_factory import backend_factory

__all__ = ['MarginalizedGraphKernel']


def flatten(iterable):
    for item in iterable:
        if hasattr(item, '__iter__'):
            yield from flatten(item)
        else:
            yield item


def fold_like(flat, example):
    folded = []
    for item in example:
        if hasattr(item, '__iter__'):
            n = len(list(flatten(item)))
            folded.append(fold_like(flat[:n], item))
            flat = flat[n:]
        else:
            folded.append(flat[0])
            flat = flat[1:]
    return tuple(folded)


class MarginalizedGraphKernel:
    """Implements the random walk-based graph similarity kernel as proposed in:
    Kashima, H., Tsuda, K., & Inokuchi, A. (2003).
    Marginalized kernels between labeled graphs. *In Proceedings of the 20th
    international conference on machine learning (ICML-03)* (pp. 321-328).

    Parameters
    ----------
    node_kernel: base kernel or composition of base kernels
        A kernelet that computes the similarity between individual nodes
    edge_kernel: base kernel or composition of base kernels
        A kernelet that computes the similarity between individual edge
    kwargs: optional arguments
        p: functor or 'uniform' or 'default'
            The starting probability of the random walk on each node. Must be
            either a functor that takes in a node (a dataframe row) and returns
            a number, or the name of a built-in distribution. Currently, only
            'uniform' and 'default' are implemented.
            Note that a custom probability does not have to be normalized.
        q: float in (0, 1)
            The probability for the random walk to stop during each step.
    """
    trait_t = namedtuple(
        'Traits', 'diagonal, symmetric, nodal, lmin, eval_gradient'
    )

    @classmethod
    def traits(cls, diagonal=False, symmetric=False, nodal=False, lmin=0,
               eval_gradient=False):
        traits = cls.trait_t(
            diagonal, symmetric, nodal, lmin, eval_gradient
        )
        if traits.eval_gradient is True:
            if nodal is not False:
                raise ValueError(
                    'Gradients can only be evaluated with nodal=False'
                )
        return traits

    def __init__(self, node_kernel, edge_kernel, p='default', q=0.01,
                 q_bounds=(1e-4, 1 - 1e-4), dtype=np.float, backend='auto'):
        self.node_kernel = node_kernel
        self.edge_kernel = edge_kernel
        self.p = self._get_starting_probability(p)
        self.q = q
        self.q_bounds = q_bounds
        self.element_dtype = dtype

        self.backend = backend_factory(backend)

    def _get_starting_probability(self, p):
        if isinstance(p, str):
            if p == 'uniform' or p == 'default':
                return lambda i, n: 1.0
            else:
                raise ValueError('Unknown starting probability distribution %s'
                                 % self.p)
        else:
            return p

    def __call__(self, X, Y=None, eval_gradient=False, nodal=False, lmin=0,
                 timing=False):
        """Compute pairwise similarity matrix between graphs

        Parameters
        ----------
        X: list of N graphs
            The graphs must all have same node and edge attributes.
        Y: None or list of M graphs
            The graphs must all have same node and edge attributes.
        nodal: bool
            If True, return node-wise similarities; otherwise, return graphwise
            similarities.
        lmin: 0 or 1
            Number of steps to skip in each random walk path before similarity
            is computed.
            lmin + 1 corresponds to the starting value of l in the summation
            of Eq. 1 in Tang & de Jong, 2019 https://doi.org/10.1063/1.5078640
            (or the first unnumbered equation as in Kashima, Tsuda, and
            Inokuchi, 2003).

        Returns
        -------
        numpy.array
            if Y is None, return a square matrix containing pairwise
            similarities between the graphs in X; otherwise, returns a matrix
            containing similarities across graphs in X and Y.
        """
        timer = Timer()
        backend = self.backend
        traits = self.traits(
            symmetric=Y is None,
            nodal=nodal,
            lmin=lmin,
            eval_gradient=eval_gradient
        )

        ''' assert graph attributes are compatible with each other '''
        pred_or_tuple = Graph.has_unified_types(X + Y if Y is not None else X)
        if pred_or_tuple is not True:
            group, first, second = pred_or_tuple
            raise TypeError(
                f'The two graphs have mismatching {group} attributes or '
                'attribute types. If the attributes match in name but differ '
                'in type, try `Graph.unify_datatype` as an automatic fix.\n'
                f'First graph: {first}\n'
                f'Second graph: {second}\n'
            )

        ''' generate jobs '''
        timer.tic('generating jobs')
        if traits.symmetric:
            i, j = np.triu_indices(len(X))
            i, j = i.astype(np.uint32), j.astype(np.uint32)
        else:
            i, j = np.indices((len(X), len(Y)), dtype=np.uint32)
            j += len(X)
        jobs = backend.array(
            np.column_stack((i.ravel(), j.ravel()))
            .ravel()
            .view(np.dtype([('i', np.uint32), ('j', np.uint32)]))
        )
        timer.toc('generating jobs')

        ''' create output buffer '''
        timer.tic('creating output buffer')
        if traits.symmetric:
            starts = backend.zeros(len(X) + 1, dtype=np.uint32)
            if traits.nodal is True:
                sizes = np.array([len(g.nodes) for g in X], dtype=np.uint32)
                np.cumsum(sizes, out=starts[1:])
                n_nodes_X = int(starts[-1])
                output_shape = (n_nodes_X, n_nodes_X)
            else:
                starts[:] = np.arange(len(X) + 1)
                output_shape = (len(X), len(X))
        else:
            starts = backend.zeros(len(X) + len(Y) + 1, dtype=np.uint32)
            if traits.nodal is True:
                sizes = np.array([len(g.nodes) for g in X]
                                 + [len(g.nodes) for g in Y],
                                 dtype=np.uint32)
                np.cumsum(sizes, out=starts[1:])
                n_nodes_X = int(starts[len(X)])
                starts[len(X):] -= n_nodes_X
                n_nodes_Y = int(starts[-1])
                output_shape = (n_nodes_X, n_nodes_Y)
            else:
                starts[:len(X)] = np.arange(len(X))
                starts[len(X):] = np.arange(len(Y) + 1)
                output_shape = (len(X), len(Y))
        if traits.eval_gradient is True:
            output_shape = (*output_shape, 1 + self.n_dims)
        output = backend.empty(int(np.prod(output_shape)), np.float32)
        timer.toc('creating output buffer')

        ''' call GPU kernel '''
        timer.tic('calling GPU kernel (overall)')
        backend(
            np.concatenate((X, Y)) if Y is not None else X,
            self.node_kernel,
            self.edge_kernel,
            self.p,
            self.q,
            jobs,
            starts,
            output,
            output_shape,
            traits,
            timer,
        )
        timer.toc('calling GPU kernel (overall)')

        ''' collect result '''
        timer.tic('collecting result')
        output = output.reshape(*output_shape, order='F')
        timer.toc('collecting result')

        if timing:
            timer.report(unit='ms')
        timer.reset()

        if traits.eval_gradient is True:
            return (
                output[:, :, 0].astype(self.element_dtype),
                output[:, :, 1:].astype(self.element_dtype)
            )
        else:
            return output.astype(self.element_dtype)

    def diag(self, X, nodal=False, lmin=0, timing=False):
        """Compute the self-similarities for a list of graphs

        Parameters
        ----------
        X: list of N graphs
            The graphs must all have same node attributes and edge attributes.
        nodal: bool
            If True, returns a vector containing nodal self similarties; if
            False, returns a vector containing graphs' overall self
            similarities; if 'block', return a list of square matrices which
            forms a block-diagonal matrix, where each diagonal block represents
            the pairwise nodal similarities within a graph.
        lmin: 0 or 1
            Number of steps to skip in each random walk path before similarity
            is computed.
            lmin + 1 corresponds to the starting value of l in the summation
            of Eq. 1 in Tang & de Jong, 2019 https://doi.org/10.1063/1.5078640
            (or the first unnumbered equation as in Kashima, Tsuda, and
            Inokuchi, 2003).

        Returns
        -------
        numpy.array or list of np.array(s)
            If nodal=True, returns a vector containing nodal self similarties;
            if nodal=False, returns a vector containing graphs' overall self
            similarities; if nodal = 'block', return a list of square matrices,
            each being a pairwise nodal similarity matrix within a graph.
        """
        timer = Timer()
        backend = self.backend

        ''' assert graph attributes are compatible with each other '''
        pred_or_tuple = Graph.has_unified_types(X)
        if pred_or_tuple is not True:
            group, first, second = pred_or_tuple
            raise TypeError(
                f'The two graphs have mismatching {group} attributes or '
                'attribute types.'
                'If the attribute names do match, then try to unify data '
                'types automatically with `Graph.unify_datatype`.\n'
                f'First graph: {first}\n'
                f'Second graph: {second}\n'
            )

        ''' generate jobs '''
        timer.tic('generating jobs')
        i = np.arange(len(X), dtype=np.uint32)
        jobs = backend.array(
            np.column_stack((i, i))
            .ravel()
            .view(np.dtype([('i', np.uint32), ('j', np.uint32)]))
        )
        timer.toc('generating jobs')

        ''' create output buffer '''
        timer.tic('creating output buffer')
        starts = backend.zeros(len(X) + 1, dtype=np.uint32)
        if nodal is True:
            sizes = np.array([len(g.nodes) for g in X], dtype=np.uint32)
            np.cumsum(sizes, out=starts[1:])
            output_length = int(starts[-1])
        elif nodal is False:
            starts[:] = np.arange(len(X) + 1)
            output_length = len(X)
        elif nodal == 'block':
            sizes = np.array([len(g.nodes) for g in X], dtype=np.uint32)
            np.cumsum(sizes**2, out=starts[1:])
            output_length = int(starts[-1])
        else:
            raise(ValueError("Invalid 'nodal' option '%s'" % nodal))
        output = backend.empty(output_length, np.float32)
        timer.toc('creating output buffer')

        ''' call GPU kernel '''
        timer.tic('calling GPU kernel (overall)')
        backend(
            X,
            self.node_kernel,
            self.edge_kernel,
            self.p,
            self.q,
            jobs,
            starts,
            output,
            (output_length, 1),
            self.traits(diagonal=True, nodal=nodal, lmin=lmin),
            timer,
        )
        timer.toc('calling GPU kernel (overall)')

        ''' post processing '''
        timer.tic('collecting result')
        if nodal == 'block':
            output = [output[s:s + n**2].reshape(n, n)
                      for s, n in zip(starts[:-1], sizes)]
        timer.toc('collecting result')

        if timing:
            timer.report(unit='ms')
        timer.reset()

        return output.astype(self.element_dtype)

    """⭣⭣⭣⭣⭣ scikit-learn interoperability methods ⭣⭣⭣⭣⭣"""

    def is_stationary(self):
        return False

    @property
    def requires_vector_input(self):
        return False

    @property
    def n_dims(self):
        '''p.theta + q + node_kernel.theta + edge_kernel.theta'''
        return len(self.theta)

    @property
    def hyperparameters(self):
        return namedtuple(
            'GraphKernelHyperparameters',
            ['q', 'node', 'edge']
        )(self.q,
          self.node_kernel.theta,
          self.edge_kernel.theta)

    @property
    def theta(self):
        return np.log(np.fromiter(flatten(self.hyperparameters), np.float))

    @theta.setter
    def theta(self, value):
        (self.q,
         self.node_kernel.theta,
         self.edge_kernel.theta
         ) = fold_like(np.exp(value), self.hyperparameters)

    @property
    def hyperparameter_bounds(self):
        return namedtuple(
            'GraphKernelHyperparameterBounds',
            ['q', 'node', 'edge']
        )(self.q_bounds,
          self.node_kernel.bounds,
          self.edge_kernel.bounds)

    @property
    def bounds(self):
        return np.log(np.fromiter(flatten(self.hyperparameter_bounds),
                                  np.float)).reshape(-1, 2, order='C')

    def clone_with_theta(self, theta):
        clone = copy.deepcopy(self)
        clone.theta = theta
        return clone

    def get_params(self, deep=False):
        return dict(
            node_kernel=self.node_kernel,
            edge_kernel=self.edge_kernel,
            p=self.p,
            q=self.q,
            q_bounds=self.q_bounds,
            backend=self.backend
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from ._backend import Backend
from ._backend_cuda import CUDABackend


def backend_factory(backend, *args, **kwargs):
    if isinstance(backend, Backend):
        return backend
    elif backend == 'cuda':
        return CUDABackend(*args, **kwargs)
    elif backend == 'auto':
        try:
            return CUDABackend(*args, **kwargs)
        except Exception as e:
            raise RuntimeError(f'Cannot auto-select backend: {e}')
    else:
        raise ValueError(f'Unknown backend {backend}')
        raise ValueError('Unknown backend {}'.format(backend))
